"""Testing!"""

import pytest
from pyiem.nws import product, ugc
from pyiem.nws.product import (
    WMO_RE,
    TextProductException,
    date_tokens2datetime,
    checker,
    str2polygon,
)
from pyiem.nws.products import parser as productparser
from pyiem.util import utc, get_test_file


def test_tropical_channels():
    """Test the channels we have some products go into."""
    prod = productparser(get_test_file("tropical/TCUCP1.txt"))
    j = prod.get_jabbers("")
    assert "TCUCP" in j[0][2]["channels"].split(",")


def test_first_flapping():
    """Test for product crossing month backwards, prevent flapping test."""
    # Scenario is utcnow is the first and the WMO header is > 25th
    utcnow = utc(2012, 6, 1)
    prod = productparser(get_test_file("SVRBMX.txt"), utcnow=utcnow)
    assert prod.valid == utc(2012, 5, 31, 23, 11)


def test_no_afos():
    """Test product without AFOS/AWIPS ID."""
    with pytest.raises(TextProductException):
        productparser(get_test_file("PMDSA.txt").replace("PMDSA ", ""))


def test_str2polygon():
    """Test our str2polygon implementation."""
    res = str2polygon("4400 3200 4500 3300 4400 3300")
    assert res is not None


def test_checker():
    """Test that exceptions are raised in certain cases."""
    with pytest.raises(TextProductException):
        checker(-90, 91, "")
    with pytest.raises(TextProductException):
        checker(-650, 31, "")


def test_datetokens_just_hour():
    """Test that we can handle having just an hour."""
    tokens = ["3", "PM", "CDT", "", "MAR", "20", "2019"]
    z, tz, valid = date_tokens2datetime(tokens)
    assert z == "CDT"
    local = valid.astimezone(tz)
    assert local.hour == 15


def test_datetokens():
    """Test that we can rectify a bad hour value."""
    tokens = ["18:45", "PM", "CDT", "", "MAR", "20", "2019"]
    z, tz, valid = date_tokens2datetime(tokens)
    assert z == "CDT"
    local = valid.astimezone(tz)
    assert local.hour == 18


def test_wpc():
    """Can we deal with the AWIPS ID in WPC products."""
    tp = productparser(get_test_file("PMDSA.txt"))
    res = tp.get_jabbers("http://localhost")
    ans = (
        "WBC issues South America Forecast Discussion (PMD) at Jul 13, "
        "11:17 AM EDT http://localhost?pid=202007131517-KWBC-FXSA20-PMDSA"
    )
    assert res[0][0] == ans


def test_200913_dualtime():
    """Process a HLS in two timezones, sigh."""
    tp = productparser(get_test_file("HLS.txt"))
    assert not tp.warnings
    assert tp.z == "EDT"


def test_200731_cvt():
    """See that we handle CVT timezone products."""
    tp = productparser(get_test_file("TCDAT5_CVT.txt"))
    res = tp.get_jabbers("http://localhost")
    ans = (
        "NHC issues TCDAT5 (TCD) at Jul 31, 8:00 PM CVT "
        "http://localhost?pid=202007312100-KNHC-WTNT45-TCDAT5"
    )
    assert res[0][0] == ans


def test_200731_bogus_timezone():
    """Test that we don't bomb out with unknown timezone."""
    tp = productparser(
        get_test_file("TCDAT5_CVT.txt").replace(" CVT ", " ZZT ")
    )
    res = tp.get_jabbers("http://localhost")
    ans = (
        "NHC issues TCDAT5 (TCD) at Jul 31, 8:00 PM ZZT "
        "http://localhost?pid=202007312000-KNHC-WTNT45-TCDAT5"
    )
    assert res[0][0] == ans
    assert len(tp.warnings) == 1
    assert tp.warnings[0] == "product timezone 'ZZT' unknown"


def test_180321_mst():
    """Do we do the right thing with MST products whilst in DST"""
    tp = productparser(get_test_file("AFDMST.txt"))
    res = tp.get_jabbers("http://localhost")
    ans = (
        "PSR issues Area Forecast Discussion (AFD) at Mar 21, 5:15 AM "
        "MST http://localhost?pid=201803211215-KPSR-FXUS65-AFDPSR"
    )
    assert res[0][0] == ans


def test_180130_chst():
    """Whoa, our offset for CHST appears to be wrong"""
    tp = productparser(get_test_file("CHST.txt"))
    res = utc(2018, 1, 30, 20, 12)
    assert tp.valid == res


def test_170411_fakemnd():
    """This RTP has a quasi-faked timestamp in the header causing error"""
    utcnow = utc(2017, 4, 10, 23, 50)
    tp = productparser(get_test_file("RTPSGX.txt"), utcnow=utcnow)
    res = utc(2017, 4, 10, 23, 30)
    assert tp.valid == res
    res = utc(2017, 4, 10, 23, 24)
    assert tp.wmo_valid == res


def test_151024_cae():
    """Make sure this CAE product works and does not throw an UGC error"""
    tp = productparser(get_test_file("CAEIA.txt"))
    assert tp.afos == "CAEIA"


def test_resent():
    """ Make sure we can tell a ...RESENT product """
    tp = productparser(get_test_file("MWWBRO.txt"))
    assert tp.is_resent()


def test_wmoheader():
    """" Make sure we can handle some header variations """
    ar = [
        "FTUS43 KOAX 102320    ",
        "FTUS43 KOAX 102320  COR ",
        "FTUS43 KOAX 102320  COR  ",
        "FTUS43 KOAX 102320",
    ]
    for item in ar:
        assert WMO_RE.match(item) is not None


def test_rfd():
    """ Parse a RFD """
    tp = productparser(get_test_file("RFDOAX.txt"))
    assert tp.get_channels()[0] == "RFDOAX"
    j = tp.get_jabbers("http://localhost")
    ans = (
        "OAX issues Grassland Fire Danger "
        "(RFD) at Jan 19, 4:10 AM CST ...MODERATE FIRE DANGER TODAY... "
        "http://localhost?pid=201501191010-KOAX-FNUS63-RFDOAX"
    )
    assert j[0][0] == ans


def test_hwo():
    """ Parse a HWO """
    tp = productparser(get_test_file("HWO.txt"))
    assert tp.get_channels()[0] == "HWOLOT"
    j = tp.get_jabbers("http://localhost")
    ans = (
        "LOT issues Hazardous Weather Outlook "
        "(HWO) at Jan 8, 3:23 PM CST "
        "http://localhost?pid=201301082123-KLOT-FLUS43-HWOLOT"
    )
    assert j[0][0] == ans


def test_140710_wmoheader_fail():
    """ Make sure COR in WMO header does not trip us up"""
    tp = product.TextProduct(get_test_file("MANANN.txt"))
    assert tp.afos == "MANANN"
    assert tp.is_correction()


def test_now_jabber():
    """ See if we can process a NOW and get the jabber result """
    tp = product.TextProduct(get_test_file("NOWDMX.txt"))
    j = tp.get_jabbers("http://localhost")
    ans = (
        "DMX issues Short-term Forecast (NOW) at Mar 4, 8:42 AM CST "
        "http://localhost?pid=201003041442-KDMX-FPUS73-NOWDMX"
    )
    assert j[0][0] == ans


def test_nomnd_with_timestamp():
    """ Make sure we process timestamps correctly when there is no MND"""
    utcnow = utc(2013, 12, 31, 18)
    tp = product.TextProduct(get_test_file("MAVWC0.txt"), utcnow=utcnow)
    ts = utc(2014, 1, 1)
    assert tp.valid == ts


def test_empty():
    """ see what happens when we send a blank string """
    with pytest.raises(TextProductException):
        product.TextProduct("")


def test_invalid_mnd_date():
    """ Check parsing of timestamp  """
    answer = utc(2013, 1, 3, 6, 16)
    tp = product.TextProduct(get_test_file("CLI/CLINYC.txt"))
    assert tp.valid == answer


def test_ugc_error130214():
    """ Check parsing of SPSJAX  """
    tp = product.TextProduct(get_test_file("SPSJAX.txt"))
    assert tp.segments[0].ugcs, [
        ugc.UGC("FL", "Z", 23),
        ugc.UGC("FL", "Z", 25),
        ugc.UGC("FL", "Z", 30),
        ugc.UGC("FL", "Z", 31),
        ugc.UGC("FL", "Z", 32),
    ]


def test_no_ugc():
    """ Product that does not have UGC encoding """
    data = get_test_file("CCFMOB.txt")
    tp = product.TextProduct(data)
    assert not tp.segments[0].ugcs


def test_ugc_invalid_coding():
    """ UGC code regression """
    data = get_test_file("FLW_badugc.txt")
    tp = product.TextProduct(data)
    # self.assertRaises(ugc.UGCParseException, product.TextProduct, data )
    assert not tp.segments[0].ugcs


def test_000000_ugctime():
    """ When there is 000000 as UGC expiration time """
    tp = product.TextProduct(get_test_file("RECFGZ.txt"))
    assert tp.segments[0].ugcexpire is None


def test_stray_space_in_ugc():
    """ When there are stray spaces in the UGC! """
    tp = product.TextProduct(get_test_file("RVDCTP.txt"))
    assert len(tp.segments[0].ugcs) == 28


def test_ugc_in_hwo():
    """ Parse UGC codes in a HWO """
    tp = product.TextProduct(get_test_file("HWO.txt"))
    assert tp.segments[1].ugcs == [
        ugc.UGC("LM", "Z", 740),
        ugc.UGC("LM", "Z", 741),
        ugc.UGC("LM", "Z", 742),
        ugc.UGC("LM", "Z", 743),
        ugc.UGC("LM", "Z", 744),
        ugc.UGC("LM", "Z", 745),
    ]


def test_afos():
    """ check AFOS PIL Parsing """
    tp = product.TextProduct(get_test_file("AFD.txt"))
    assert tp.afos == "AFDBOX"


def test_source():
    """ check tp.source Parsing """
    tp = product.TextProduct(get_test_file("AFD.txt"))
    assert tp.source == "KBOX"


def test_wmo():
    """ check tp.wmo Parsing """
    tp = product.TextProduct(get_test_file("AFD.txt"))
    assert tp.wmo == "FXUS61"


def test_notml():
    """ check TOR without TIME...MOT...LOC """
    tp = product.TextProduct(get_test_file("TOR.txt"))
    assert tp.segments[0].tml_dir is None


def test_signature():
    """ check svs_search """
    tp = product.TextProduct(get_test_file("TOR.txt"))
    assert tp.get_signature() == "CBD"


def test_spanishMWW():
    """ check spanish MWW does not break things """
    tp = product.TextProduct(get_test_file("MWWspanish.txt"))
    assert tp.z is None


def test_svs_search():
    """ check svs_search """
    tp = product.TextProduct(get_test_file("TOR.txt"))
    ans = (
        "* AT 1150 AM CDT...THE NATIONAL WEATHER SERVICE "
        "HAS ISSUED A TORNADO WARNING FOR DESTRUCTIVE "
        "WINDS OVER 110 MPH IN THE EYE WALL AND INNER RAIN "
        "BANDS OF HURRICANE KATRINA. THESE WINDS WILL "
        "OVERSPREAD MARION...FORREST AND LAMAR COUNTIES "
        "DURING THE WARNING PERIOD."
    )
    assert tp.segments[0].svs_search() == ans


def test_product_id():
    """ check valid Parsing """
    tp = product.TextProduct(get_test_file("AFD.txt"))
    assert tp.get_product_id() == "201211270001-KBOX-FXUS61-AFDBOX"


def test_valid():
    """ check valid Parsing """
    tp = product.TextProduct(get_test_file("AFD.txt"))
    ts = utc(2012, 11, 27, 0, 1)
    assert tp.valid == ts


def test_FFA():
    """ check FFA Parsing """
    tp = product.TextProduct(get_test_file("FFA.txt"))
    assert tp.segments[0].get_hvtec_nwsli() == "NWYI3"


def test_valid_nomnd():
    """ check valid (no Mass News) Parsing """
    utcnow = utc(2012, 11, 27)
    tp = product.TextProduct(get_test_file("AFD_noMND.txt"), utcnow=utcnow)
    ts = utc(2012, 11, 27, 0, 1)
    assert tp.valid == ts


def test_headlines():
    """ check headlines Parsing """
    tp = product.TextProduct(get_test_file("AFDDMX.txt"))
    ans = [
        "UPDATED FOR 18Z AVIATION DISCUSSION",
        "Bogus second line with a new line",
    ]
    assert tp.segments[0].headlines == ans


def test_tml():
    """ Test TIME...MOT...LOC parsing """
    ts = utc(2012, 5, 31, 23, 10)
    tp = product.TextProduct(get_test_file("SVRBMX.txt"))
    assert tp.segments[0].tml_dir == 238
    assert tp.segments[0].tml_valid == ts
    assert tp.segments[0].tml_sknt == 39
    assert tp.segments[0].tml_giswkt == "SRID=4326;POINT(-88.53 32.21)"


def test_bullets():
    """ Test bullets parsing """
    tp = product.TextProduct(get_test_file("TORtag.txt"))
    assert len(tp.segments[0].bullets) == 4
    ans = (
        "LOCATIONS IMPACTED INCLUDE... MARYSVILLE...LOVILIA"
        "...HAMILTON AND BUSSEY."
    )
    assert tp.segments[0].bullets[3] == ans

    tp = product.TextProduct(get_test_file("FLSDMX.txt"))
    assert len(tp.segments[2].bullets) == 7
    ans = (
        "IMPACT...AT 35.5 FEET...WATER AFFECTS 285TH "
        "AVENUE NEAR SEDAN BOTTOMS...OR JUST EAST OF THE "
        "INTERSECTION OF 285TH AVENUE AND 570TH STREET."
    )
    assert tp.segments[2].bullets[6] == ans


def test_tags():
    """ Test tags parsing """
    tp = product.TextProduct(get_test_file("TORtag.txt"))
    assert tp.segments[0].tornadotag == "OBSERVED"
    assert tp.segments[0].tornadodamagetag == "SIGNIFICANT"


def test_longitude_processing():
    """ Make sure that parsed longitude values are negative! """
    tp = product.TextProduct(get_test_file("SVRBMX.txt"))
    assert abs(tp.segments[0].sbw.exterior.xy[0][0] - -88.39) < 0.01


def test_giswkt():
    """ Test giswkt parsing """
    tp = product.TextProduct(get_test_file("SVRBMX.txt"))
    assert abs(tp.segments[0].sbw.area - 0.16) < 0.01

    ans = (
        "SRID=4326;MULTIPOLYGON "
        "(((-88.390000 32.590000, -88.130000 32.760000, "
        "-88.080000 32.720000, -88.110000 32.690000, "
        "-88.040000 32.690000, -88.060000 32.640000, "
        "-88.080000 32.640000, -88.060000 32.590000, "
        "-87.930000 32.630000, -87.870000 32.570000, "
        "-87.860000 32.520000, -87.920000 32.520000, "
        "-87.960000 32.470000, -88.030000 32.430000, "
        "-88.050000 32.370000, -87.970000 32.350000, "
        "-87.940000 32.310000, -88.410000 32.310000, "
        "-88.390000 32.590000)))"
    )
    assert tp.segments[0].giswkt == ans



"""Unit tests for pyiem.nws.vtec"""

from pyiem.util import utc
from pyiem.nws import vtec

EX1 = "/O.NEW.KJAN.TO.W.0130.050829T1651Z-050829T1815Z/"


def test_beginstring_combos():
    """Test our begin string logic."""
    vc = vtec.parse("/O.NEW.KJAN.TO.W.0130.000000T0000Z-000000T0000Z/")
    assert vc[0].get_begin_string(None) == ""


def test_endstring_until_further_notice():
    """Make sure that the end time string is empty for cancel action"""
    vc = vtec.parse("/O.NEW.KJAN.TO.W.0130.050829T1651Z-000000T0000Z/")
    assert vc[0].get_end_string(None) == "until further notice"


def test_badformat():
    """Test that we handle bad timestamps."""
    res = vtec.contime("AABBCCTHHMMZ")
    assert res is None


def test_fireweather():
    """Do we return different things for FW"""
    res = vtec.get_ps_string("FW", "A")
    assert res == "Fire Weather Watch"
    res = vtec.get_ps_string("FW", "W")
    assert res == "Red Flag Warning"


def test_s3():
    """Test our s3 API."""
    vc = vtec.parse(EX1)
    assert vc[0].s3() == "TO.W.130"


def test_s2():
    """Test our s2 API."""
    vc = vtec.parse(EX1)
    assert vc[0].s2() == "TO.W"


def test_get_id():
    """check that getID() works as we expect"""
    vc = vtec.parse(EX1)
    assert vc[0].get_id(2005) == "2005-KJAN-TO-W-0130"


def test_endstring():
    """Make sure that the end time string is empty for cancel action"""
    vc = vtec.parse("/O.CAN.KJAN.TO.W.0130.050829T1651Z-050829T1815Z/")
    assert vc[0].get_end_string(None) == ""


def test_begints():
    """ check vtec.begints Parsing """
    vc = vtec.parse(EX1)
    ts = utc(2005, 8, 29, 16, 51)
    assert vc[0].begints == ts


def test_endts():
    """ check vtec.endts Parsing """
    vc = vtec.parse(EX1)
    ts = utc(2005, 8, 29, 18, 15)
    assert vc[0].endts == ts


def test_product_string():
    """ check vtec.product_string() formatting """
    vc = vtec.parse("/O.NEW.KJAN.TO.W.0130.050829T1651Z-050829T1815Z/")
    assert vc[0].product_string() == "issues Tornado Warning"



"""National Weather Service Location Idenitifiers (NWSLI)

A class to store metadata associated with a NWSLI entry.
"""
from shapely.geometry import Point


class NWSLI:
    """National Weather Service Location Idenitifiers (NWSLI)"""

    def __init__(self, identifier, name=None, wfos=None, lon=0, lat=0):
        """Constructor

        Args:
          identifier(str): The string identifier for this NWSLI
          name(str, optional): The free-form text name of this location
          wfo(list, optional): The wfo(s) associated with this NWSLI
          lon(float, optional): The longitude in decimal degrees
          lat(float, optional): The latitude in decimal degress
        """
        self.id = identifier
        self.name = name
        self.wfos = wfos if wfos is not None else []
        self.geometry = Point([lon, lat])

    def get_name(self):
        """Return the name of this NWSLI, uses `self.id` if name is unset

        Returns:
            str: the name of this site
        """
        if self.name is None:
            return f"(({self.id}))"
        return self.name



"""Test TAF Parsing"""

from pyiem.util import get_test_file, utc
from pyiem.nws.products import parser as tafparser


def test_parse():
    """TAF type"""
    utcnow = utc(2017, 7, 25)
    prod = tafparser(get_test_file("TAF/TAFJFK.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://localhost", "http://localhost")
    ans = (
        "OKX issues Terminal Aerodrome Forecast (TAF) "
        "at Jul 25, 13:41 UTC for JFK http://localhost?"
        "pid=201707251341-KOKX-FTUS41-TAFJFK"
    )
    assert j[0][0] == ans
    assert "TAFJFK" in j[0][2]["channels"].split(",")



"""Serialization of geometries for use in pyIEM.plot mapping

We use a pickled protocol=2, which is compat binary.
"""
import datetime

import pickle
from shapely.wkb import loads
from pyiem.util import get_dbconn

PATH = "../src/pyiem/data"
# Be annoying
print("Be sure to run this against Mesonet database and not laptop!")


def dump_states(fn):
    """states."""
    pgconn = get_dbconn("postgis", user="nobody")
    cursor = pgconn.cursor()

    cursor.execute(
        """ SELECT state_abbr,
    ST_Simplify(the_geom, 0.01),
    ST_x(ST_Centroid(the_geom)), ST_Y(ST_Centroid(the_geom)) from states"""
    )

    data = {}
    for row in cursor:
        data[row[0]] = dict(
            geom=loads(row[1], hex=True), lon=row[2], lat=row[3]
        )
        # for polygon in geom:
        #    data[row[0]].append(np.asarray(polygon.exterior))

    with open("%s/%s" % (PATH, fn), "wb") as f:
        pickle.dump(data, f, 2)


def dump_climdiv(fn):
    """climate divisions."""
    pgconn = get_dbconn("postgis", user="nobody")
    cursor = pgconn.cursor()

    cursor.execute(
        """ SELECT iemid, geom,
    ST_x(ST_Centroid(geom)), ST_Y(ST_Centroid(geom))
    from climdiv"""
    )

    data = {}
    for row in cursor:
        data[row[0]] = dict(
            geom=loads(row[1], hex=True), lon=row[2], lat=row[3]
        )
        # for polygon in geom:
        #    data[row[0]].append(np.asarray(polygon.exterior))

    with open("%s/%s" % (PATH, fn), "wb") as f:
        pickle.dump(data, f, 2)


def dump_cwa(fn):
    """WFOs."""
    pgconn = get_dbconn("postgis", user="nobody")
    cursor = pgconn.cursor()

    cursor.execute(
        """ SELECT wfo, ST_Simplify(the_geom, 0.01),
    ST_x(ST_Centroid(the_geom)), ST_Y(ST_Centroid(the_geom)), region
    from cwa"""
    )

    data = {}
    for row in cursor:
        data[row[0]] = dict(
            geom=loads(row[1], hex=True), lon=row[2], lat=row[3], region=row[4]
        )
    with open("%s/%s" % (PATH, fn), "wb") as f:
        pickle.dump(data, f, 2)


def dump_iowawfo(fn):
    """ A region with the Iowa WFOs"""
    pgconn = get_dbconn("postgis", user="nobody")
    cursor = pgconn.cursor()

    cursor.execute(
        """ SELECT ST_Simplify(ST_Union(the_geom), 0.01)
        from cwa
        WHERE wfo in ('DMX', 'ARX', 'DVN', 'OAX', 'FSD')"""
    )
    row = cursor.fetchone()

    geo = loads(row[0], hex=True)
    data = dict()
    data["iowawfo"] = dict(geom=geo, lon=geo.centroid.x, lat=geo.centroid.y)
    with open("%s/%s" % (PATH, fn), "wb") as f:
        pickle.dump(data, f, 2)


def dump_ugc(gtype, fn):
    """UGCS."""
    pgconn = get_dbconn("postgis", user="nobody")
    cursor = pgconn.cursor()

    cursor.execute(
        """ SELECT ugc, wfo, simple_geom,
        ST_x(centroid), ST_Y(centroid) from ugcs
        WHERE end_ts is null and substr(ugc, 3, 1) = %s""",
        (gtype,),
    )

    data = {}
    for row in cursor:
        data[row[0]] = dict(
            cwa=row[1][:3],
            geom=loads(row[2], hex=True),
            lon=row[3],
            lat=row[4],
        )
        # for polygon in geom:
        #    data[row[0]].append(np.asarray(polygon.exterior))

    with open("%s/%s" % (PATH, fn), "wb") as f:
        pickle.dump(data, f, 2)


def check_file(fn):
    """regression check."""
    sts = datetime.datetime.now()
    data = pickle.load(open("%s/%s" % (PATH, fn), "rb"))
    ets = datetime.datetime.now()

    print(
        "runtime: %.5fs, entries: %s, fn: %s"
        % ((ets - sts).total_seconds(), len(data.keys()), fn)
    )


def main():
    """Go Main"""
    dump_iowawfo("iowawfo.pickle")
    dump_ugc("C", "ugcs_county.pickle")
    dump_ugc("Z", "ugcs_zone.pickle")
    check_file("ugcs_county.pickle")
    check_file("ugcs_zone.pickle")
    dump_cwa("cwa.pickle")
    check_file("cwa.pickle")
    dump_climdiv("climdiv.pickle")
    check_file("climdiv.pickle")
    dump_states("us_states.pickle")
    check_file("us_states.pickle")


if __name__ == "__main__":
    main()



"""Custom Exceptions."""


class NoDataFound(Exception):
    """Exception for when no data was found for request."""


class UnitsError(Exception):
    """Exception for bad Units."""


class TextProductException(Exception):
    """Exception for Text Parsing."""


class HWOException(Exception):
    """Exception for HWO Parsing."""


class MCDException(Exception):
    """ Exception """


class NHCException(Exception):
    """ Exception """


class SAWException(Exception):
    """Custom local exception"""


class UGCParseException(Exception):
    """Custom Exception this parser can raise"""


class CLIException(Exception):
    """Custom Exception for CLI Parsing Issues"""



"""Space Weather Center processor"""

import pyiem.nws.product as product


class SpaceWxProduct(product.TextProduct):
    """ Class for parsing and representing Space Wx Products """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        product.TextProduct.__init__(
            self, text, utcnow, ugc_provider, nwsli_provider
        )
        self.title = "Unknown (AWIPSID: %s)" % (self.afos,)
        self.parse_title()

    def parse_title(self):
        """ Figure out the title of this product """
        if len(self.sections) < 2:
            return
        self.title = self.sections[2].split("\n")[0]

    def get_jabbers(self, uri, _uri2=None):
        """ Custom Implementation of the TextProduct#get_jabbers """
        url = "%s?pid=%s" % (uri, self.get_product_id())
        xtra = {
            "channels": "WNP,%s" % (self.afos,),
            "twitter": "SWPC issues %s %s" % (self.title, url),
        }
        plain = ("Space Weather Prediction Center issues %s %s") % (
            self.title,
            url,
        )
        html = (
            "<p>Space Weather Prediction Center "
            '<a href="%s">issues %s</a></p>'
        ) % (url, self.title)
        return [(plain, html, xtra)]


def parser(buf, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """ A parser implementation """
    return SpaceWxProduct(buf, utcnow, ugc_provider, nwsli_provider)



from __future__ import absolute_import
from pyiem.nws.product import TextProduct, TextProductException, WMO_RE, AFOSRE
from . import spacewx
from . import cli
from . import hwo
from . import lsr
from . import mcd
from . import nhc
from . import sps
from . import taf


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """Omnibus parser of NWS Text Data

    This is intended to be a catch-all parser of text data.  As it currently
    stands, it does not correctly hand products off to the correct
    sub-processor, but some day it will!

    Args:
      text (str): The actual product text, this can have the <cntr>-a
        character to start the string.
      utcnow (datetime, optional): What is the current time, this is useful
        for when ingesting old data.  Many times, the product does not contain
        enough information to assign a current valid timestamp to it.  So we
        need to know the current timestamp to do the relative computation.
      ugc_provider (dict, optional): Provides NWS UGC metadata, the dictionary
        keys are UGC codes.
      nwsli_provider (dict, optional): Provides NWS Location Identifiers to
        allow lookup of geographic information for station identifiers.

    Returns:
      TextProduct: A TextProduct instance

    """

    tmp = text[:100].replace("\r\r\n", "\n")
    m = WMO_RE.search(tmp)
    if m is not None:
        d = m.groupdict()
        if d["cccc"] == "KWNP":
            return spacewx.parser(text, utcnow, ugc_provider, nwsli_provider)

    tokens = AFOSRE.findall(tmp)
    if not tokens:
        raise TextProductException("Could not locate AFOS Identifier")
    afos = tokens[0]
    if afos[:3] == "CLI":
        return cli.parser(text, utcnow, ugc_provider, nwsli_provider)
    elif afos[:3] == "TCP":
        return nhc.parser(text, utcnow, ugc_provider, nwsli_provider)
    elif afos[:3] == "HWO":
        return hwo.parser(text, utcnow, ugc_provider, nwsli_provider)
    elif afos in ["SWOMCD", "FFGMPD"]:
        return mcd.parser(text, utcnow, ugc_provider, nwsli_provider)
    elif afos[:3] == "LSR":
        return lsr.parser(text, utcnow, ugc_provider, nwsli_provider)
    elif afos[:3] == "TAF":
        return taf.parser(text, utcnow, ugc_provider, nwsli_provider)
    elif afos[:3] == "SPS":
        return sps.parser(text, utcnow, ugc_provider, nwsli_provider)

    return TextProduct(text, utcnow, ugc_provider, nwsli_provider)



"""TAF Parsing"""

from pyiem.nws.product import TextProduct
from pyiem import reference


class TAFProduct(TextProduct):
    """
    Represents a TAF
    """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        TextProduct.__init__(self, text, utcnow, ugc_provider, nwsli_provider)

    def get_channels(self):
        """ Return a list of channels """
        return [self.afos, "TAF...", f"{self.source}.TAF"]

    def get_jabbers(self, uri, _uri2=None):
        """ Get the jabber variant of this message """
        res = []
        # These products can be ignored
        if self.afos is None:
            return res
        url = f"{uri}?pid={self.get_product_id()}"
        aaa = self.afos[:3]
        nicedate = self.get_nicedate()
        plain = ("%s issues %s (%s) at %s for %s %s") % (
            self.source[1:],
            reference.prodDefinitions.get(aaa, aaa),
            aaa,
            nicedate,
            self.afos[3:],
            url,
        )
        html = ('<p>%s issues <a href="%s">%s (%s)</a> at %s for %s</p>') % (
            self.source[1:],
            url,
            reference.prodDefinitions.get(aaa, aaa),
            aaa,
            nicedate,
            self.afos[3:],
        )
        xtra = {
            "channels": ",".join(self.get_channels()),
            "product_id": self.get_product_id(),
            "twitter": plain,
        }
        res.append((plain, html, xtra))
        return res


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """ Helper function """
    return TAFProduct(text, utcnow, ugc_provider, nwsli_provider)



"""Support NWS VTEC encoding"""
import re
from datetime import timezone, timedelta, datetime

from pyiem.util import LOG

VTEC_RE = (
    r"(/([A-Z])\.([A-Z]+)\.([A-Z]+)\.([A-Z]+)\.([A-Z])\."
    r"([0-9]+)\.([0-9TZ]+)-([0-9TZ]+)/)"
)

VTEC_CLASS = {
    "O": "Operational",
    "T": "Test",
    "E": "Experimental",
    "X": "Experimental VTEC",
}

VTEC_ACTION = {
    "NEW": "issues",
    "CON": "continues",
    "EXA": "expands area to include",
    "EXT": "extends time of",
    "EXB": "extends time and expands area to include",
    "UPG": "issues upgrade to",
    "CAN": "cancels",
    "EXP": "expires",
    "ROU": "routine",
    "COR": "corrects",
}

VTEC_SIGNIFICANCE = {
    "W": "Warning",
    "Y": "Advisory",
    "A": "Watch",
    "S": "Statement",
    "O": "Outlook",
    "N": "Synopsis",
    "F": "Forecast",
}

# https://www.nws.noaa.gov/directives/sym/pd01017003curr.pdf
VTEC_PHENOMENA = {
    "AF": "Ashfall",
    "AS": "Air Stagnation",
    "BH": "Beach Hazard",
    "BS": "Blowing Snow",
    "BW": "Brisk Wind",
    "BZ": "Blizzard",
    "CF": "Coastal Flood",
    "DF": "Debris Flow",
    "DS": "Dust Storm",
    "DU": "Blowing Dust",
    "EC": "Extreme Cold",
    "EH": "Excessive Heat",
    "EW": "Extreme Wind",
    "FA": "Flood",
    "FF": "Flash Flood",
    "FG": "Dense Fog",
    "FL": "Flood",
    "FR": "Frost",
    "FW": "Red Flag",
    "FZ": "Freeze",
    "GL": "Gale",
    "HF": "Hurricane Force Wind",
    "HI": "Inland Hurricane",
    "HS": "Heavy Snow",
    "HT": "Heat",
    "HU": "Hurricane",
    "HW": "High Wind",
    "HY": "Hydrologic",
    "HZ": "Hard Freeze",
    "IP": "Sleet",
    "IS": "Ice Storm",
    "LB": "Lake Effect Snow and Blowing Snow",
    "LE": "Lake Effect Snow",
    "LO": "Low Water",
    "LS": "Lakeshore Flood",
    "LW": "Lake Wind",
    "MA": "Marine",
    "MF": "Marine Dense Fog",
    "MH": "Marine Ashfall",
    "MS": "Marine Dense Smoke",
    "RB": "Small Craft for Rough",
    "RP": "Rip Currents",
    "SB": "Snow and Blowing",
    "SC": "Small Craft",
    "SE": "Hazardous Seas",
    "SI": "Small Craft for Winds",
    "SM": "Dense Smoke",
    "SN": "Snow",
    "SQ": "Snow Squall",
    "SR": "Storm",
    "SS": "Storm Surge",
    "SU": "High Surf",
    "SV": "Severe Thunderstorm",
    "SW": "Small Craft for Hazardous Seas",
    "TI": "Inland Tropical Storm",
    "TO": "Tornado",
    "TR": "Tropical Storm",
    "TS": "Tsunami",
    "TY": "Typhoon",
    "UP": "Ice Accretion",
    "WC": "Wind Chill",
    "WI": "Wind",
    "WS": "Winter Storm",
    "WW": "Winter Weather",
    "ZF": "Freezing Fog",
    "ZR": "Freezing Rain",
    "ZY": "Freezing Spray",
}

# Taken from http://www.weather.gov/help-map
NWS_COLORS = {
    "AS.Y": "#808080",
    "AF.Y": "#696969",
    "AF.W": "#A9A9A9",
    "BH.S": "#40E0D0",
    "BZ.W": "#FF4500",
    "BZ.A": "#ADFF2F",
    "DU.Y": "#BDB76B",
    "BW.Y": "#D8BFD8",
    "CF.Y": "#7CFC00",
    "CF.S": "#6B8E23",
    "CF.W": "#228B22",
    "CF.A": "#66CDAA",
    "DS.W": "#FFE4C4",
    "FG.Y": "#708090",
    "SM.Y": "#F0E68C",
    "DU.W": "#FFE4C4",
    "EH.W": "#C71585",
    "EH.Y": "#800000",
    "EC.W": "#0000FF",
    "EC.A": "#0000FF",
    "EW.W": "#FF8C00",
    "FA.Y": "#00FF7F",
    "FW.A": "#FFDEAD",
    "FF.S": "#8B0000",
    "FF.W": "#8B0000",
    "FF.A": "#2E8B57",
    "FL.Y": "#00FF7F",
    "FL.S": "#00FF00",
    "FL.W": "#00FF00",
    "FL.A": "#2E8B57",
    "FZ.W": "#483D8B",
    "FZ.A": "#00FFFF",
    "ZF.Y": "#008080",
    "ZR.Y": "#DA70D6",
    "FR.Y": "#6495ED",
    "GL.W": "#DDA0DD",
    "GL.A": "#FFC0CB",
    "HZ.W": "#9400D3",
    "HZ.A": "#4169E1",
    "SE.W": "#D8BFD8",
    "SE.A": "#483D8B",
    "HT.Y": "#FF7F50",
    "SU.Y": "#BA55D3",
    "SU.W": "#228B22",
    "HW.W": "#DAA520",
    "HW.A": "#B8860B",
    "HF.W": "#CD5C5C",
    "HF.A": "#9932CC",
    "HU.W": "#DC143C",
    "HU.A": "#FF00FF",
    "HY.Y": "#00FF7F",
    "IS.W": "#8B008B",
    "LE.Y": "#48D1CC",
    "LE.W": "#008B8B",
    "LE.A": "#87CEFA",
    "LW.Y": "#D2B48C",
    "LS.Y": "#7CFC00",
    "LS.S": "#6B8E23",
    "LS.W": "#228B22",
    "LS.A": "#66CDAA",
    "LO.Y": "#A52A2A",
    "MA.S": "#FFDAB9",
    "FW.W": "#FF1493",
    "RP.S": "#40E0D0",
    "SQ.W": "#C71585",
    "SV.W": "#FFA500",
    "SV.A": "#DB7093",
    "SC.Y": "#D8BFD8",
    "SW.Y": "#D8BFD8",
    "RB.Y": "#D8BFD8",
    "SI.Y": "#D8BFD8",
    "MA.W": "#FFA500",
    "TO.W": "#FF0000",
    "TO.A": "#FFFF00",
    "TR.S": "#FFE4B5",
    "TR.W": "#B22222",
    "TR.A": "#F08080",
    "TS.Y": "#D2691E",
    "TS.W": "#FD6347",
    "TS.A": "#FF00FF",
    "TY.W": "#DC143C",
    "TY.A": "#FF00FF",
    "WI.Y": "#D2B48C",
    "WC.Y": "#AFEEEE",
    "WC.W": "#B0C4DE",
    "WC.A": "#5F9EA0",
    "WS.W": "#FF69B4",
    "WS.A": "#4682B4",
    "WW.Y": "#7B68EE",
}


def parse(text):
    """ I look for and return vtec objects as I find them """
    vtec = []
    tokens = re.findall(VTEC_RE, text)
    for token in tokens:
        vtec.append(VTEC(token))
    return vtec


def contime(text):
    """Convert text into a UTC datetime."""
    # The 0000 is the standard VTEC undefined time
    if text.startswith("0000"):
        return None
    try:
        ts = datetime.strptime(text, "%y%m%dT%H%MZ")
    except Exception as err:
        LOG.exception(err)
        return None
    # NWS has a bug sometimes whereby 1969 or 1970s timestamps are emitted
    if ts.year < 1971:
        return None
    return ts.replace(tzinfo=timezone.utc)


def get_ps_string(phenomena, significance):
    """Return the combination of Phenomena + Significance as string"""
    pstr = VTEC_PHENOMENA.get(phenomena, f"Unknown {phenomena}")
    astr = VTEC_SIGNIFICANCE.get(significance, f"Unknown {significance}")
    # Hack for special FW case
    if significance == "A" and phenomena == "FW":
        pstr = "Fire Weather"
    return f"{pstr} {astr}"


def get_action_string(action):
    """Return the action string"""
    return VTEC_ACTION.get(action, f"unknown {action}")


class VTEC:
    """A single VTEC encoding instance"""

    def __init__(self, tokens):
        self.line = tokens[0]
        self.status = tokens[1]
        self.action = tokens[2]
        self.office = tokens[3][1:]
        self.office4 = tokens[3]
        self.phenomena = tokens[4]
        self.significance = tokens[5]
        self.etn = int(tokens[6])
        self.begints = contime(tokens[7])
        self.endts = contime(tokens[8])
        # Not explicitly defined, but set later by product parsing logic
        self.year = None

    def s3(self):
        """Return a commonly used string representation."""
        return f"{self.phenomena}.{self.significance}.{self.etn}"

    def s2(self):
        """Return a commonly used string representation."""
        return f"{self.phenomena}.{self.significance}"

    def get_end_string(self, prod):
        """ Return an appropriate end string for this VTEC """
        if self.action in ["CAN", "EXP"]:
            return ""
        if self.endts is None:
            return "until further notice"
        fmt = "%b %-d, %-I:%M %p"
        if self.endts < (prod.valid + timedelta(hours=1)):
            fmt = "%-I:%M %p"
        if prod.tz is None:
            fmt = "%b %-d, %-H:%M"
        localts = self.endts
        if prod.tz is not None:
            localts = self.endts.astimezone(prod.tz)
        # A bit of complexity as offices may not implement daylight saving
        if prod.z is not None and prod.z.endswith("ST") and localts.dst():
            localts -= timedelta(hours=1)
        return "till %s %s" % (
            localts.strftime(fmt),
            prod.z if prod.z is not None else "UTC",
        )

    def get_begin_string(self, prod):
        """Return an appropriate beginning string for this VTEC"""
        if self.begints is None:
            return ""
        fmt = "%b %-d, %-I:%M %p"
        if self.begints < (prod.valid + timedelta(hours=1)):
            fmt = "%-I:%M %p"
        localts = self.begints.astimezone(prod.tz)
        # A bit of complexity as offices may not implement daylight saving
        if prod.z.endswith("ST") and localts.dst():
            localts -= timedelta(hours=1)
        return "valid at %s %s" % (localts.strftime(fmt), prod.z)

    def url(self, year):
        """ Generate a VTEC url string needed """
        return ("%s-%s-%s-%s-%s-%s-%04i") % (
            year if self.year is None else self.year,
            self.status,
            self.action,
            self.office4,
            self.phenomena,
            self.significance,
            self.etn,
        )

    def get_id(self, year):
        """Return a custom string identifier for this VTEC product

        This is used by the Live client
        """
        return "%s-%s-%s-%s-%04i" % (
            year if self.year is None else self.year,
            self.office4,
            self.phenomena,
            self.significance,
            self.etn,
        )

    def __str__(self):
        """Return string representation"""
        return self.line

    def get_ps_string(self):
        """Return the combination of Phenomena + Significance as string"""
        return get_ps_string(self.phenomena, self.significance)

    def get_action_string(self):
        """Return the action string"""
        return get_action_string(self.action)

    def product_string(self):
        """Return the combination of action and phenomena+significance"""
        return "%s %s" % (self.get_action_string(), self.get_ps_string())



"""Test coverage for our jinja2 templates."""

from pyiem.templates.iem import TEMPLATE, get_template


def test_basic_usage():
    """Test that we can import what we think we can."""
    res = TEMPLATE.render()
    assert res is not None


def test_get_template():
    """See that we can load other templates."""
    tpl = get_template("app.j2")
    assert tpl is not None



"""
module twistedpg.py
Author: Federico Di Gregorio
http://twistedmatrix.com/pipermail/twisted-python/2006-April/012955.html
"""

from psycopg2 import connect
from psycopg2 import connect as _2connect
from psycopg2.extensions import connection as _2connection
from psycopg2.extras import DictCursor

del connect


def connect(*args, **kwargs):
    """Proxy connect with an additional kwarg set"""
    kwargs["connection_factory"] = connection
    return _2connect(*args, **kwargs)


class connection(_2connection):
    def cursor(self):
        # DictCursor allows for both integer and key based row access
        return _2connection.cursor(self, cursor_factory=DictCursor)



"""Test NLDN."""

import pytest
from pyiem.nws.products.nldn import parser
from pyiem.util import get_test_file, get_dbconn


@pytest.fixture
def cursor():
    """Return a database cursor."""
    return get_dbconn("nldn").cursor()


def test_1_basic():
    """CLIBNA is a new diction"""
    np = parser(get_test_file("NLDN/example.bin", fponly=True))
    assert len(np.df.index) == 50


def test_sql(cursor):
    """Test that we can insert data."""
    np = parser(get_test_file("NLDN/example.bin", fponly=True))
    np.sql(cursor)



"""Helper for PRISM BIL data"""
import datetime

import numpy as np

NX = 1405
NY = 621

# This is the center of the upper left pixel
NORTH = 49.92
SOUTH = 24.04
WEST = -125.0
EAST = -66.46
DX = 0.0417
DY = 0.0417

XAXIS = WEST + np.arange(NX) * DX
YAXIS = SOUTH + np.arange(NY) * DY


def daily_offset(ts):
    """ Compute the timestamp index in the netcdf file """
    # In case ts is passed here as a datetime.date object
    ts = datetime.datetime(ts.year, ts.month, ts.day)
    base = ts.replace(
        month=1, day=1, hour=0, minute=0, second=0, microsecond=0
    )
    days = (ts - base).days
    return int(days)


def find_ij(lon, lat):
    """Compute which grid cell this lon, lat resides within"""
    if lon < WEST or lon >= EAST or lat < SOUTH or lat >= NORTH:
        return None, None

    i = np.digitize([lon], XAXIS)[0] - 1
    j = np.digitize([lat], YAXIS)[0] - 1

    return i, j



"""Massive omnibus of testing for pyiem.nws.products."""
# pylint: disable=redefined-outer-name

import pytest
import psycopg2.extras
from pyiem.nws.products import parser
from pyiem.util import get_dbconn, utc, get_test_file


@pytest.fixture
def dbcursor():
    """Return a disposable database cursor."""
    pgconn = get_dbconn("postgis")
    return pgconn.cursor(cursor_factory=psycopg2.extras.DictCursor)


def filter_warnings(ar, startswith="get_gid"):
    """Remove non-deterministic warnings

    Some of our tests produce get_gid() warnings, which are safe to ignore
    for the purposes of this testing
    """
    return [a for a in ar if not a.startswith(startswith)]


def test_201124_tab_in_afos():
    """Test that we can deal with a tab in the AFOS ID."""
    data = get_test_file("FFW/FFWGUM.txt").replace("FFWGUM", "FFWGUM\t")
    prod = parser(data)
    assert prod.afos == "FFWGUM"


def test_181207_issue74_guam():
    """Guam's longitudes are east, not west like code assumes."""
    prod = parser(get_test_file("FFW/FFWGUM.txt"))
    ans = "SRID=4326;MULTIPOLYGON (((145.800000 15.160000, 145.740000"
    assert prod.segments[0].giswkt.startswith(ans)


def test_180917_issue63_tweet_length():
    """Make sure this tweet text is not too long!"""
    utcnow = utc(2018, 9, 15, 11, 56)
    prod = parser(get_test_file("LSR/LSRCRP.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://iem.local/")
    assert j[0][2]["twitter"] == (
        "At 6:45 AM CDT, 2 NNE Odem [San Patricio Co, TX] DEPT OF HIGHWAYS "
        "reports FLOOD. ROAD CLOSURE FM 1944 BETWEEN U.S. HIGHWAY 77 AND "
        "SODVILLE ROAD (TEXAS DEPARTMENT OF TRANSPORATION. DRIVETEXAS.ORG.) "
        "LATITUDE/LONGITUDE MARKS APPROXIMATE POSITION OF FM 194... "
        "http://iem.local/#CRP/201809151145/201809151145"
    )


def test_170116_mixedlsr():
    """LSRBOU has mixed case, see what we can do"""
    utcnow = utc(2016, 11, 29, 22, 00)
    prod = parser(get_test_file("mIxEd_CaSe/LSRBOU.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://iem.local/")
    assert j[0][2]["twitter"] == (
        "At 11:00 AM MST, Akron [Washington Co, CO] ASOS reports "
        "High Wind of M63 MPH "
        "http://iem.local/#BOU/201611291800/201611291800"
    )


def test_180710_issue58():
    """Crazy MST during MDT"""
    utcnow = utc(2018, 7, 9, 22, 59)
    prod = parser(get_test_file("LSR/LSRPSR.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://iem.local/")
    ans = (
        "At 3:57 PM MST, 5 WNW Florence [Pinal Co, AZ] TRAINED SPOTTER "
        "reports FLASH FLOOD. STREET FLOODING WITH WATER OVER THE CURBS "
        "IN THE MERRILL RANCH DEVELOPMENT OF FLORENCE. "
        "http://iem.local/#PSR/201807092257/201807092257"
    )
    assert j[0][2]["twitter"] == ans
    ans = (
        "5 WNW Florence [Pinal Co, AZ] TRAINED SPOTTER reports FLASH FLOOD "
        "at 3:57 PM MST -- STREET FLOODING WITH WATER OVER THE CURBS IN "
        "THE MERRILL RANCH DEVELOPMENT OF FLORENCE. "
        "http://iem.local/#PSR/201807092257/201807092257"
    )
    assert j[0][0] == ans


def test_180705_iembot_issue9():
    """LSRBOU has mixed case, see what we can do"""
    utcnow = utc(2018, 7, 4, 22, 11)
    prod = parser(get_test_file("LSR/LSRDMX.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://iem.local/")
    assert j[0][2]["twitter"] == (
        "At 1:30 PM CDT, 1 WNW Lake Mills [Winnebago Co, IA] TRAINED SPOTTER "
        "reports TSTM WND GST of E61 MPH. SPOTTER MEASURED 61 MPH WIND GUST. "
        "HIS CAR DOOR WAS ALSO CAUGHT BY THE WIND WHEN HE WAS OPENING "
        "THE DOOR, PUSHING THE DOOR INTO HIS FACE. THIS CONTACT"
        "... http://iem.local/#DMX/201807041830/201807041830"
    )


def test_171026_mixedlsr():
    """LSRBYZ has mixed case, see what we can do"""
    utcnow = utc(2017, 10, 29, 19, 18)
    prod = parser(get_test_file("mIxEd_CaSe/LSRBYZ.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://iem.local/")
    assert j[0][2]["twitter"] == (
        "[Delayed Report] On Oct 26, at 1:00 AM MDT, 3 SSW Luther "
        "[Carbon Co, MT] Mesonet reports Snow of 1.00 inch "
        "http://iem.local/#BYZ/201710260700/201710260700"
    )


def test_170419_tcp_mixedcase():
    """Mixed case TCP1"""
    prod = parser(get_test_file("TCPAT1_mixedcase.txt"))
    j = prod.get_jabbers("")
    assert j


def test_170403_badtime():
    """Handle when a colon is added to a timestamp"""
    prod = parser(get_test_file("FLWBOI.txt"))
    prod.get_jabbers("http://localhost", "http://localhost")
    ans = utc(2017, 4, 2, 2, 30)
    assert prod.valid == ans


def test_170324_badformat():
    """Look into exceptions"""
    utcnow = utc(2017, 3, 22, 2, 35)
    prod = parser(get_test_file("LSR/LSRPIH.txt"), utcnow=utcnow)
    prod.get_jabbers("http://iem.local/")
    assert len(prod.warnings) == 2
    assert not prod.lsrs


def test_170324_ampersand():
    """LSRs with ampersands may cause trouble"""
    utcnow = utc(2015, 12, 29, 18, 23)
    prod = parser(get_test_file("LSR/LSRBOXamp.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://iem.local/")
    ans = (
        "Lunenberg [Worcester Co, MA] HAM RADIO reports SNOW of 2.00 INCH "
        "at 11:36 AM EST -- HAM RADIO AND THIS DARYL ADDED &amp; and &lt; "
        "and &gt; http://iem.local/#BOX/201512291636/201512291636"
    )
    assert j[0][0] == ans


def test_170207_mixedhwo():
    """Check our parsing of mixed case HWO"""
    prod = parser(get_test_file("mIxEd_CaSe/HWOLOT.txt"))
    j = prod.get_jabbers("http://iem.local/")
    assert not prod.warnings
    assert len(j[0]) == 3


def test_160618_chst_tz():
    """Product has timezone of ChST, do we support it?"""
    prod = parser(get_test_file("AFDPQ.txt"))
    assert prod.valid == utc(2016, 6, 18, 19, 27)


def test_151229_badgeo_lsr():
    """Make sure we reject a bad Geometry LSR"""
    utcnow = utc(2015, 12, 29, 18, 23)
    prod = parser(get_test_file("LSR/LSRBOX.txt"), utcnow=utcnow)
    assert len(prod.warnings) == 1
    assert not prod.lsrs


def test_150422_tornadomag():
    """LSRTAE see what we do with tornado magitnudes"""
    utcnow = utc(2015, 4, 22, 15, 20)
    prod = parser(get_test_file("LSR/LSRTAE.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://iem.local/")
    assert j[0][1] == (
        "<p>[Delayed Report] 4 W Bruce [Walton Co, FL] NWS EMPLOYEE "
        '<a href="http://iem.local/#TAE/201504191322/201504191322">'
        "reports TORNADO of EF0</a> at 19 Apr, 9:22 AM EDT -- "
        "SHORT EF0 TORNADO PATH CONFIRMED BY NWS DUAL POL RADAR DEBRIS "
        "SIGNATURE IN A RURAL AREA WEST OF BRUCE. DAMAGE LIKELY CONFINED "
        "TO TREES. ESTIMATED DURATION 3 MINUTES. PATH LENGTH "
        "APPROXIMATELY 1 MILE.</p>"
    )
    assert j[0][2]["twitter"] == (
        "[Delayed Report] On Apr 19, at 9:22 AM EDT, "
        "4 W Bruce [Walton Co, FL] NWS EMPLOYEE reports "
        "TORNADO of EF0. SHORT EF0 TORNADO PATH CONFIRMED BY NWS DUAL "
        "POL RADAR DEBRIS SIGNATURE IN A RURAL AREA WEST OF BRUCE. "
        "DAMAGE LIKELY CONFINED TO TREES. ESTIMATED... "
        "http://iem.local/#TAE/201504191322/201504191322"
    )


def test_150202_hwo():
    """HWORNK emitted a poorly worded error message"""
    prod = parser(get_test_file("HWORNK.txt"))
    with pytest.raises(Exception):
        prod.get_jabbers("http://localhost", "http://localhost")


def test_160418_hwospn():
    """Make sure a spanish HWO does not trip us up..."""
    utcnow = utc(2016, 4, 18, 10, 10)
    prod = parser(get_test_file("HWOSPN.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://localhost", "http://localhost")
    ans = (
        "JSJ issues Hazardous Weather Outlook (HWO) "
        "at Apr 18, 10:18 UTC http://localhost?"
        "pid=201604181018-TJSJ-FLCA42-HWOSPN"
    )
    assert j[0][0] == ans


def test_tcp():
    """ See what we can do with TCP """
    prod = parser(get_test_file("TCPAT1.txt"))
    j = prod.get_jabbers("http://localhost", "http://localhost")
    ans = (
        "National Hurricane Center issues "
        "ADVISORY 19 for POST-TROPICAL CYCLONE ARTHUR "
        "http://localhost?pid=201407051500-KNHC-WTNT31-TCPAT1"
    )
    assert j[0][0] == ans
    ans = (
        "Post-Tropical Cyclone "
        "#Arthur ADVISORY 19 issued. Strong winds and heavy rains to "
        "continue over portions of southeastern canada through tonight "
        "http://go.usa.gov/W3H"
    )
    assert j[0][2]["twitter"] == ans


def test_140820_badtimestamp():
    """ Check our invalid timestamp exception and how it is written """
    with pytest.raises(Exception):
        parser(get_test_file("RWSGTF_badtime.txt"))


def test_160904_resent():
    """Is this product a correction?"""
    prod = parser(get_test_file("TCVAKQ.txt"))
    jmsgs = prod.get_jabbers("http://localhost")
    ans = (
        "AKQ issues Tropical Watch/Warning Local Statement (TCV) "
        "at Sep 2, 11:55 AM EDT ...TROPICAL STORM WARNING IN EFFECT... "
        "http://localhost?pid=201609021555-KAKQ-WTUS81-TCVAKQ"
    )
    assert jmsgs[0][2]["twitter"] == ans


def test_jabber_lsrtime():
    """Make sure delayed LSRs have proper dates associated with them"""
    utcnow = utc(2014, 6, 6, 16)
    prod = parser(get_test_file("LSR/LSRFSD.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://iem.local")
    ans = (
        "<p>[Delayed Report] 2 SSE Harrisburg [Lincoln Co, SD] "
        'TRAINED SPOTTER <a href="http://iem.local#FSD/201406052040/'
        '201406052040">reports TORNADO</a> at 5 Jun, 3:40 PM CDT -- '
        "ON GROUND ALONG HIGHWAY 11 NORTH OF 275TH ST</p>"
    )
    assert j[0][1] == ans


def test_spacewx():
    """See if we can parse a space weather product """
    utcnow = utc(2014, 5, 10)
    prod = parser(get_test_file("SPACEWX.txt"), utcnow=utcnow)
    j = prod.get_jabbers("http://localhost/")
    ans = (
        "Space Weather Prediction Center issues "
        "CANCEL WATCH: Geomagnetic Storm Category G3 Predicted "
        "http://localhost/?pid=201405101416-KWNP-WOXX22-WATA50"
    )
    assert j[0][0] == ans


def test_140522_blowingdust():
    """Make sure we can deal with invalid LSR type """
    prod = parser(get_test_file("LSR/LSRTWC.txt"))
    assert not prod.lsrs


def test_01():
    """LSR.txt process a valid LSR without blemish """
    utcnow = utc(2013, 7, 23, 23, 54)
    prod = parser(get_test_file("LSR/LSR.txt"), utcnow=utcnow)
    assert prod.lsrs[0].remark is None
    assert len(prod.lsrs) == 58

    assert abs(prod.lsrs[57].magnitude_f - 73) < 0.01
    assert prod.lsrs[57].county == "MARION"
    assert prod.lsrs[57].state == "IA"
    assert abs(prod.lsrs[57].get_lon() - -93.11) < 0.01
    assert abs(prod.lsrs[57].get_lat() - 41.3) < 0.01

    assert prod.is_summary()
    assert prod.lsrs[57].wfo == "DMX"

    answer = utc(2013, 7, 23, 3, 55)
    assert prod.lsrs[57].valid == answer
    j = prod.get_jabbers("http://iem.local/")
    assert j[57][0], (
        "Knoxville Airport [Marion Co, IA] AWOS reports NON-TSTM WND "
        "GST of M73 MPH at 22 Jul, 10:55 PM CDT -- HEAT BURST. "
        "TEMPERATURE ROSE FROM 70 TO 84 IN 15 MINUTES AND DEW POINT "
        "DROPPED FROM 63 TO 48 IN 10 MINUTES. "
        "http://iem.local/#DMX/201307230355/201307230355"
    )

    with pytest.deprecated_call():
        prod.lsrs[5].tweet()



"""Make sure our METAR parsing works!"""
# pylint: disable=redefined-outer-name

import pytest
import psycopg2.extras
from pyiem.reference import TRACE_VALUE
from pyiem.nws.products import metarcollect
from pyiem.util import get_dbconn, utc, get_test_file

PARSER = metarcollect.parser
NWSLI_PROVIDER = {
    "CYYE": dict(network="CA_BC_ASOS"),
    "QQQQ": dict(network="FAKE", tzname="America/Chicago"),
    "SPS": dict(wfo="OUN"),
    "MIA": dict(wfo="MIA"),
    "ALO": dict(wfo="DSM"),
    "EST": dict(wfo="EST"),
}
metarcollect.JABBER_SITES = {"KALO": None}


@pytest.fixture
def dbcursor():
    """Return a disposable database cursor."""
    pgconn = get_dbconn("iem")
    cursor = pgconn.cursor(cursor_factory=psycopg2.extras.DictCursor)
    # Create fake station, so we can create fake entry in summary
    # and current tables
    cursor.execute(
        """
        INSERT into stations(id, network, iemid, tzname)
        VALUES ('QQQQ', 'FAKE', -1, 'America/Chicago')
    """
    )
    cursor.execute(
        """
        INSERT into current(iemid, valid) VALUES
        (-1, '2015-09-01 00:00+00')
    """
    )
    cursor.execute(
        """
        INSERT into summary_2015(iemid, day) VALUES
        (-1, '2015-09-01')
    """
    )
    return cursor


def test_corrected(dbcursor):
    """Test that the COR does not get dropped from the raw METAR."""
    code = (
        "KSAT 121751Z COR VRB03KT 10SM SCT043 32/19 A3002 RMK AO2 SLP144 "
        "T03220194 10328 20228 58011 $="
    )
    mtr = metarcollect.METARReport(code, year=2020, month=10)
    iemob, _ = mtr.to_iemaccess(dbcursor)
    assert mtr.code == iemob.data["raw"]


def test_issue92_6hour(dbcursor):
    """Can we get the 6 hour right."""
    utcnow = utc(2015, 9, 1, 23)
    header = "000 \r\r\nSAUS44 KISU 011200\r\r\nMETAR "
    # 4 PM temp is 63
    data = header + (
        "QQQQ 012153Z 23016G25KT 10SM FEW025 SCT055 17/04 A2982 RMK "
        "AO2 SLP104 T01720044=\r\r\n"
    )
    prod = PARSER(data, utcnow=utcnow, nwsli_provider=NWSLI_PROVIDER)
    iemob, _ = prod.metars[0].to_iemaccess(dbcursor)
    iemob.load(dbcursor)
    assert iemob.data["max_tmpf"] == 63
    # 5 PM temp is 61, but has 6 hour low of 54, high of 64
    data = header + (
        "QQQQ 012253Z 27005KT 10SM FEW025 BKN055 16/05 A2982 RMK AO2 "
        "SLP105 T01610050 10178 20122 56014=\r\r\n"
    )
    prod = PARSER(data, utcnow=utcnow, nwsli_provider=NWSLI_PROVIDER)
    iemob, _ = prod.metars[0].to_iemaccess(dbcursor)
    iemob.load(dbcursor)
    assert iemob.data["max_tmpf"] == 64
    assert iemob.data["min_tmpf"] == 54


def test_issue92_6hour_nouse(dbcursor):
    """We should not use the 6 hour in this case."""
    utcnow = utc(2015, 9, 1, 9)
    header = "000 \r\r\nSAUS44 KISU 011200\r\r\nMETAR "
    # 1 AM temp is 63
    data = header + (
        "QQQQ 010653Z 23016G25KT 10SM FEW025 SCT055 17/04 A2982 RMK "
        "AO2 SLP104 T01720044=\r\r\n"
    )
    prod = PARSER(data, utcnow=utcnow, nwsli_provider=NWSLI_PROVIDER)
    iemob, _ = prod.metars[0].to_iemaccess(dbcursor)
    iemob.load(dbcursor)
    assert iemob.data["max_tmpf"] == 63
    # 2 AM temp is 61, but has 6 hour low of 54, high of 64
    data = header + (
        "QQQQ 010753Z 27005KT 10SM FEW025 BKN055 16/05 A2982 RMK AO2 "
        "SLP105 T01610050 10178 20122 56014=\r\r\n"
    )
    prod = PARSER(data, utcnow=utcnow, nwsli_provider=NWSLI_PROVIDER)
    iemob, _ = prod.metars[0].to_iemaccess(dbcursor)
    iemob.load(dbcursor)
    assert iemob.data["max_tmpf"] == 63
    assert iemob.data["min_tmpf"] == 61


def test_issue89_peakwind(dbcursor):
    """Are we roundtripping peak wind."""
    code = (
        "KALO 010001Z AUTO 17027G37KT 10SM FEW030 SCT110 19/16 A2979 RMK AO2 "
        "PK WND 18049/2025 RAE48 SLP088 P0005 60014 T01890156 58046"
    )
    mtr = metarcollect.METARReport(code, year=2017, month=1)
    iemob, _ = mtr.to_iemaccess(dbcursor)
    assert iemob.data["peak_wind_time"] == utc(2016, 12, 31, 20, 25)


def test_190118_ice(dbcursor):
    """Process a ICE Report."""
    mtr = metarcollect.METARReport(
        (
            "KABI 031752Z 30010KT 6SM BR FEW009 OVC036 02/01 A3003 RMK AO2 "
            "SLP176 60001 I1000 T00170006 10017 21006 56017"
        )
    )
    assert mtr.ice_accretion_1hr is not None
    iemob, _ = mtr.to_iemaccess(dbcursor)
    assert iemob.data["ice_accretion_1hr"] == TRACE_VALUE


def test_180604_nonascii():
    """See that we don't error on non-ASCII METARs"""
    utcnow = utc(2018, 6, 4)
    prod = PARSER(get_test_file("METAR/badchars.txt"), utcnow=utcnow)
    assert len(prod.metars) == 3


def test_future():
    """Can we handle products that are around the first"""
    utcnow = utc(2017, 12, 1)
    prod = PARSER(get_test_file("METAR/first.txt"), utcnow=utcnow)
    assert len(prod.metars) == 2
    assert prod.metars[0].time.month == 11
    assert prod.metars[1].time.month == 12


def test_180201_unparsed():
    """For some reason, this collective was not parsed?!?!"""
    utcnow = utc(2018, 2, 1, 0)
    prod = PARSER(
        get_test_file("METAR/collective2.txt"),
        utcnow=utcnow,
        nwsli_provider=NWSLI_PROVIDER,
    )
    assert len(prod.metars) == 35
    assert prod.metars[0].time.month == 1


def test_170824_sa_format():
    """Don't be so noisey when we encounter SA formatted products"""
    utcnow = utc(2017, 8, 24, 14)
    prod = PARSER(
        get_test_file("METAR/sa.txt"),
        utcnow=utcnow,
        nwsli_provider=NWSLI_PROVIDER,
    )
    assert not prod.metars


def test_170809_nocrcrlf():
    """Product fails WMO parsing due to usage of RTD as bbb field"""
    utcnow = utc(2017, 8, 9, 9)
    prod = PARSER(
        get_test_file("METAR/rtd_bbb.txt"),
        utcnow=utcnow,
        nwsli_provider=NWSLI_PROVIDER,
    )
    assert len(prod.metars) == 1


def test_metarreport(dbcursor):
    """Can we do things with the METARReport"""
    utcnow = utc(2013, 8, 8, 12, 53)
    mtr = metarcollect.METARReport(
        (
            "SPECI CYYE 081253Z 01060KT 1/4SM FG SKC 10/10 A3006 RMK P0000 "
            "FG6 SLP188="
        )
    )
    assert mtr.wind_gust is None
    mtr.time = utcnow
    mtr.iemid = "CYYE"
    mtr.network = "CA_BC_ASOS"
    iemob, _ = mtr.to_iemaccess(dbcursor)
    assert iemob.data["station"] == "CYYE"
    assert iemob.data["phour"] == TRACE_VALUE
    assert mtr.wind_message() == "gust of 0 knots (0.0 mph) from N @ 1253Z"
    # can we round trip the gust
    iemob.load(dbcursor)
    assert iemob.data["gust"] is None


def test_basic(dbcursor):
    """Simple tests"""
    utcnow = utc(2013, 8, 8, 14)
    prod = PARSER(
        get_test_file("METAR/collective.txt"),
        utcnow=utcnow,
        nwsli_provider=NWSLI_PROVIDER,
    )
    assert not prod.warnings
    assert len(prod.metars) == 11
    jmsgs = prod.get_jabbers("localhost")
    assert len(jmsgs) == 6
    ans = (
        "None,None (SPS) ASOS reports Hail -- KSPS 081352Z 10015KT 10SM "
        "TSGRRA BKN022CB BKN050 BKN200 25/16 A2967 RMK AO2 TSB38RAB25GRB49 "
        "SLP036 LTGICCCCG OHD TS OHD GR 1/3 P0000 T02500161"
    )
    assert jmsgs[0][2]["twitter"] == ans

    iemob, _ = prod.metars[1].to_iemaccess(dbcursor)
    assert abs(iemob.data["phour"] - 0.46) < 0.01



# -*- coding: utf-8 -*-
# pylint: disable=import-outside-toplevel,unbalanced-tuple-unpacking
"""Utility functions for pyIEM package

This module contains utility functions used by various parts of the codebase.
"""
import os
import sys
import time
import random
import logging
from datetime import timezone, datetime
import re
import warnings
import getpass
from socket import error as socket_error

from six import string_types

# NB: some third party stuff is expensive to import, so let us be lazy

# NB: We shall not be importing other parts of pyIEM here as we then get
# circular references.

SEQNUM = re.compile(r"^[0-9]{3}\s?$")
# Setup a default logging instance for this module
LOG = logging.getLogger("pyiem")
LOG.addHandler(logging.NullHandler())


class CustomFormatter(logging.Formatter):
    """A custom log formatter class."""

    def format(self, record):
        """Return a string!"""
        return "[%s %6.3f %s:%s %s] %s" % (
            time.strftime("%H:%M:%S", time.localtime(record.created)),
            record.relativeCreated / 1000.0,
            record.filename,
            record.lineno,
            record.funcName,
            record.getMessage(),
        )


def html_escape(val):
    """Wrapper around cgi.escape depreciation."""
    from html import escape

    return escape(val)


def get_test_file(name, fponly=False):
    """Helper to get data for test usage."""
    basedir = os.path.dirname(__file__)
    fn = f"{basedir}/../../data/product_examples/{name}"
    fp = open(fn, "rb")
    if fponly:
        return fp
    return fp.read().decode("utf-8")


def logger(name="pyiem", level=None):
    """Get pyiem's logger with a stream handler attached.

    Args:
      name (str): The name of the logger to get, default pyiem
      level (logging.LEVEL): The log level for this pyiem logget, default is
        INFO for non interactive sessions, DEBUG otherwise

    Returns:
      logger instance
    """
    ch = logging.StreamHandler()
    ch.setFormatter(CustomFormatter())
    log = logging.getLogger(name)
    log.addHandler(ch)
    if level is None and sys.stdout.isatty():
        level = logging.DEBUG
    log.setLevel(level if level is not None else logging.INFO)
    return log


def find_ij(lons, lats, lon, lat):
    """Compute the i,j closest cell."""
    import numpy as np

    dist = ((lons - lon) ** 2 + (lats - lat) ** 2) ** 0.5
    (xidx, yidx) = np.unravel_index(dist.argmin(), dist.shape)
    return xidx, yidx


def get_twitter(screen_name):
    """Provide an authorized Twitter API Client

    Args:
      screen_name (str): The twitter user we are fetching creds for
    """
    import twython

    dbconn = get_dbconn("mesosite")
    cursor = dbconn.cursor()
    props = get_properties(cursor)
    # fetch the oauth saved creds
    cursor.execute(
        "select access_token, access_token_secret from iembot_twitter_oauth "
        "WHERE screen_name = %s",
        (screen_name,),
    )
    row = cursor.fetchone()
    dbconn.close()
    return twython.Twython(
        props["bot.twitter.consumerkey"],
        props["bot.twitter.consumersecret"],
        row[0],
        row[1],
    )


def ssw(mixedobj):
    """python23 wrapper for sys.stdout.write

    Args:
      mixedobj (str or bytes): what content we want to send
    """
    stdout = getattr(sys.stdout, "buffer", sys.stdout)
    if isinstance(mixedobj, string_types):
        stdout.write(mixedobj.encode("utf-8"))
    else:
        stdout.write(mixedobj)


def ncopen(ncfn, mode="r", timeout=60):
    """Safely open netcdf files

    The issue here is that we can only have the following situation for a
    given NetCDF file.
    1.  Only 1 or more readers
    2.  Only 1 appender

    The netcdf is being accessed over NFS and perhaps local disk, so writing
    lock files is problematic.

    Args:
      ncfn (str): The netCDF filename
      mode (str,optional): The netCDF4.Dataset open mode, default 'r'
      timeout (int): The total time in seconds to attempt a read, default 60

    Returns:
      `netCDF4.Dataset` or `None`
    """
    import netCDF4

    if mode != "w" and not os.path.isfile(ncfn):
        raise IOError("No such file %s" % (ncfn,))
    sts = datetime.utcnow()
    nc = None
    while (datetime.utcnow() - sts).total_seconds() < timeout:
        try:
            nc = netCDF4.Dataset(ncfn, mode)
            nc.set_auto_scale(True)
            break
        except (OSError, IOError):
            pass
        time.sleep(5)
    return nc


def utc(year=None, month=1, day=1, hour=0, minute=0, second=0, microsecond=0):
    """Create a datetime instance with tzinfo=timezone.utc

    When no arguments are provided, returns `datetime.utcnow()`.

    Returns:
      datetime with tzinfo set
    """
    if year is None:
        return datetime.utcnow().replace(tzinfo=timezone.utc)
    return datetime(
        year, month, day, hour, minute, second, microsecond
    ).replace(tzinfo=timezone.utc)


def get_dbconn(database="mesosite", user=None, host=None, port=5432, **kwargs):
    """Helper function with business logic to get a database connection

    Note that this helper could return a read-only database connection if the
    connection to the primary server fails.

    Args:
      database (str,optional): the database name to connect to.
        default: mesosite
      user (str,optional): hard coded user to connect as, default: current user
      host (str,optional): hard coded hostname to connect as,
        default: iemdb.local
      port (int,optional): the TCP port that PostgreSQL is listening
        defaults to 5432
      password (str,optional): the password to use.
      allow_failover (bool,optional): Should this method attempt to connect to
        a failover host (hard coded as iemdb2.local), default is `True`.

    Returns:
      psycopg2 database connection
    """
    import psycopg2

    if user is None:
        user = getpass.getuser()
        # We hard code the apache user back to nobody, www-data is travis-ci
        if user in ["apache", "www-data"]:
            user = "nobody"
        elif user == "akrherz":  # HACK for daryl's development, sigh
            user = "mesonet"
        elif user == "meteor_ldm":  # Another HACK
            user = "ldm"
    if host is None:
        host = "iemdb-%s.local" % (database,)
    conn_kwargs = {
        "database": database,
        "host": host,
        "user": user,
        "password": kwargs.get("password"),
        "port": port,
        "connect_timeout": kwargs.get("connect_timeout", 15),
        "gssencmode": kwargs.get("gssencmode", "disable"),
    }
    allow_failover = kwargs.pop("allow_failover", True)
    conn_kwargs.update(kwargs)
    attempt = 0
    while attempt < 3:
        attempt += 1
        try:
            return psycopg2.connect(**conn_kwargs)
        except psycopg2.ProgrammingError as exp:
            # Likely gssencmode is not permitted
            if "gssencmode" in conn_kwargs:
                conn_kwargs.pop("gssencmode")
            else:
                warnings.warn(
                    f"database connection failure: {exp}", stacklevel=2
                )
            if attempt == 3:
                raise exp
        except psycopg2.OperationalError as exp:
            # as a stop-gap, lets try connecting to iemdb2
            host2 = "iemdb2.local" if allow_failover else host
            conn_kwargs["host"] = host2
            if attempt == 3:
                raise exp
            warnings.warn(
                f"database connection failure: {exp}, trying {host2}",
                stacklevel=2,
            )


def noaaport_text(text):
    """Make whatever text look like it is NOAAPort Pristine

    Args:
      text (string): the inbound text
    Returns:
      text that looks noaaportish
    """
    # Rectify the text to remove any stray stuff
    text = text.replace("\003", "").replace("\001", "").replace("\r", "")
    # trim any right hand space
    lines = [x.rstrip() for x in text.split("\n")]
    # remove any beginning empty lines
    for pos in [0, -1]:
        while lines and lines[pos].strip() == "":
            lines.pop(pos)

    # lime 0 should be start of product sequence
    lines.insert(0, "\001")
    # line 1 should be the LDM sequence number 4 chars
    if not SEQNUM.match(lines[1]):
        if len(lines[1]) > 5:
            lines.insert(1, "000 ")
    else:
        lines[1] = f"{lines[1][:3]} "
    # last line should be the control-c, by itself
    lines.append("\003")

    return "\r\r\n".join(lines)


def get_autoplot_context(fdict, cfg):
    """Get the variables out of a dict of strings

    This helper for IEM autoplot gets values out of a dictionary of strings,
    as provided by CGI.  It does some magic to get types right, defaults right
    and so on.  The typical way this is called

        ctx = iemutils.get_context(fdict, get_description())

    Args:
      fdict (dictionary): what was likely provided by `cgi.FieldStorage()`
      cfg (dictionary): autoplot value of get_description
    Returns:
      dictionary of variable names and values, with proper types!
    """
    ctx = {}
    for opt in cfg.get("arguments", []):
        name = opt.get("name")
        default = opt.get("default")
        typ = opt.get("type")
        minval = opt.get("min")
        maxval = opt.get("max")
        optional = opt.get("optional", False)
        value = fdict.get(name)
        if optional and value is None and typ not in ["vtec_ps"]:
            continue
        if typ in ["station", "zstation", "sid", "networkselect"]:
            # A bit of hackery here if we have a name ending in a number
            netname = "network%s" % (
                name[-1] if name[-1] in ["1", "2", "3", "4", "5"] else "",
            )
            # The network variable tags along and within a non-PHP context,
            # this variable is unset, so we do some more hackery here
            ctx[netname] = fdict.get(netname, opt.get("network"))
            # Convience we load up the network metadata
            ntname = "_nt%s" % (
                name[-1] if name[-1] in ["1", "2", "3", "4", "5"] else "",
            )
            from pyiem.network import Table as NetworkTable
            from pyiem.exceptions import NoDataFound

            ctx[ntname] = NetworkTable(ctx[netname], only_online=False)
            # stations starting with _ are virtual and should not error
            if value is None:
                value = default
            if not value.startswith("_") and value not in ctx[ntname].sts:
                raise NoDataFound("Station metadata unavailable.")

        elif typ in ["int", "month", "zhour", "hour", "day", "year"]:
            if value is not None:
                value = int(value)
            if default is not None:
                default = int(default)
        elif typ == "float":
            if value is not None:
                value = float(value)
            if default is not None:
                default = float(default)
        elif typ == "select":
            options = opt.get("options", dict())
            # in case of multi, value could be a list
            if value is None:
                value = default
            elif isinstance(value, string_types):
                if value not in options:
                    value = default
                if opt.get("multiple"):
                    value = [value]
            else:
                res = []
                for subval in value:
                    if subval in options:
                        res.append(subval)
                value = res
        elif typ == "datetime":
            # tricky here, php has YYYY/mm/dd and CGI has YYYY-mm-dd
            if default is not None:
                default = datetime.strptime(default, "%Y/%m/%d %H%M")
            if minval is not None:
                minval = datetime.strptime(minval, "%Y/%m/%d %H%M")
            if maxval is not None:
                maxval = datetime.strptime(maxval, "%Y/%m/%d %H%M")
            if value is not None:
                if value.find(" ") == -1:
                    value += " 0000"
                value = datetime.strptime(value, "%Y-%m-%d %H%M")
        elif typ == "date":
            # tricky here, php has YYYY/mm/dd and CGI has YYYY-mm-dd
            if default is not None:
                default = datetime.strptime(default, "%Y/%m/%d").date()
            if minval is not None:
                minval = datetime.strptime(minval, "%Y/%m/%d").date()
            if maxval is not None:
                maxval = datetime.strptime(maxval, "%Y/%m/%d").date()
            if value is not None:
                value = datetime.strptime(value, "%Y-%m-%d").date()
        elif typ == "vtec_ps":
            # VTEC phenomena and significance
            defaults = {}
            # Only set a default value when the field is not optional
            if default is not None and not optional:
                tokens = default.split(".")
                if (
                    len(tokens) == 2
                    and len(tokens[0]) == 2
                    and len(tokens[1]) == 1
                ):
                    defaults["phenomena"] = tokens[0]
                    defaults["significance"] = tokens[1]
            for label in ["phenomena", "significance"]:
                label2 = label + name
                ctx[label2] = fdict.get(label2, defaults.get(label))
            continue
        # validation
        if minval is not None and value is not None and value < minval:
            value = default
        if maxval is not None and value is not None and value > maxval:
            value = default
        ctx[name] = value if value is not None else default
    return ctx


def exponential_backoff(func, *args, **kwargs):
    """Exponentially backoff some function until it stops erroring

    Args:
      _ebfactor (int,optional): Optional scale factor, allowing for faster test
    """
    ebfactor = float(kwargs.pop("_ebfactor", 2))
    msgs = []
    for i in range(5):
        try:
            return func(*args, **kwargs)
        except socket_error as serr:
            msgs.append("%s/5 %s traceback: %s" % (i + 1, func.__name__, serr))
            time.sleep((ebfactor ** i) + (random.randint(0, 1000) / 1000))
        except Exception as exp:
            msgs.append("%s/5 %s traceback: %s" % (i + 1, func.__name__, exp))
            time.sleep((ebfactor ** i) + (random.randint(0, 1000) / 1000))
    logging.error("%s failure", func.__name__)
    logging.error("\n".join(msgs))
    return None


def get_properties(cursor=None):
    """Fetch the properties set

    Returns:
      dict: a dictionary of property names and values (both str)
    """
    if cursor is None:
        pgconn = get_dbconn("mesosite", user="nobody")
        cursor = pgconn.cursor()
    cursor.execute("SELECT propname, propvalue from properties")
    res = {}
    for row in cursor:
        res[row[0]] = row[1]
    return res


def drct2text(drct):
    """Convert an degree value to text representation of direction.

    Args:
      drct (int or float): Value in degrees to convert to text

    Returns:
      str: String representation of the direction, could be `None`

    """
    if drct is None:
        return None
    # Convert the value into a float
    drct = float(drct)
    if drct > 360 or drct < 0:
        return None
    text = None
    if drct >= 350 or drct < 13:
        text = "N"
    elif drct < 35:
        text = "NNE"
    elif drct < 57:
        text = "NE"
    elif drct < 80:
        text = "ENE"
    elif drct < 102:
        text = "E"
    elif drct < 127:
        text = "ESE"
    elif drct < 143:
        text = "SE"
    elif drct < 166:
        text = "SSE"
    elif drct < 190:
        text = "S"
    elif drct < 215:
        text = "SSW"
    elif drct < 237:
        text = "SW"
    elif drct < 260:
        text = "WSW"
    elif drct < 281:
        text = "W"
    elif drct < 304:
        text = "WNW"
    elif drct < 324:
        text = "NW"
    else:
        text = "NNW"
    return text


def grid_bounds(lons, lats, bounds):
    """Figure out indices that we can truncate big grid

    Args:
      lons (np.array): grid lons
      lats (np.array): grid lats
      bounds (list): [x0, y0, x1, y1]

    Returns:
      [x0, y0, x1, y1]
    """
    import numpy as np

    if len(lons.shape) == 1:
        # Do 1-d work
        (x0, x1) = np.digitize([bounds[0], bounds[2]], lons)
        (y0, y1) = np.digitize([bounds[1], bounds[3]], lats)
        szx = len(lons)
        szy = len(lats)
    else:
        # Do 2-d work
        diff = ((lons - bounds[0]) ** 2 + (lats - bounds[1]) ** 2) ** 0.5
        (lly, llx) = np.unravel_index(np.argmin(diff), lons.shape)
        diff = ((lons - bounds[2]) ** 2 + (lats - bounds[3]) ** 2) ** 0.5
        (ury, urx) = np.unravel_index(np.argmin(diff), lons.shape)
        diff = ((lons - bounds[0]) ** 2 + (lats - bounds[3]) ** 2) ** 0.5
        (uly, ulx) = np.unravel_index(np.argmin(diff), lons.shape)
        diff = ((lons - bounds[2]) ** 2 + (lats - bounds[1]) ** 2) ** 0.5
        (lry, lrx) = np.unravel_index(np.argmin(diff), lons.shape)
        x0 = min([llx, ulx])
        x1 = max([lrx, urx])
        y0 = min([lry, lly])
        y1 = max([uly, ury])
        (szy, szx) = lons.shape

    return [
        int(i)
        for i in [max([0, x0]), max([0, y0]), min([szx, x1]), min([szy, y1])]
    ]



"""MCD/MPD tests."""
# pylint: disable=redefined-outer-name

import psycopg2.extras
import pytest
from pyiem.exceptions import MCDException
from pyiem.nws.products import parser
from pyiem.util import get_dbconn, utc, get_test_file


@pytest.fixture
def dbcursor():
    """Database cursor."""
    return get_dbconn("postgis").cursor(
        cursor_factory=psycopg2.extras.DictCursor
    )


def test_exceptions():
    """Test that various things lead to exceptions."""
    orig = get_test_file("MCD_MPD/SWOMCD.txt")
    with pytest.raises(MCDException):
        parser(orig.replace("LAT...LON", "BLAH"))
    # See if -40 logic works
    parser(orig.replace("44738786", "44733786"))

    prod = parser(orig.replace("CONCERNING", "BLAH"))
    assert prod.concerning is None

    prod = parser(orig.replace("VALID", "BLAH"))
    assert prod.warnings

    with pytest.raises(MCDException):
        parser(orig.replace("DISCUSSION", "BLAH"))
    with pytest.raises(MCDException):
        parser(orig.replace("ATTN...WFO", "BLAH"))


def test_datetime_coverage():
    """Test products that cross the first of the month."""
    assert parser(get_test_file("MCD_MPD/SWOMCD_sep1.txt")) is not None


def test_issue163(dbcursor):
    """Test parsing of the concerning tag."""
    prod = parser(get_test_file("MCD_MPD/SWOMCDconcerning.txt"))
    prod.database_save(dbcursor)
    dbcursor.execute(
        "SELECT concerning from mcd where product_id = %s",
        (prod.get_product_id(),),
    )
    ans = "Severe Thunderstorm Watch 60"
    assert dbcursor.fetchone()[0] == ans
    # Exercise that the remove worked.
    prod.database_save(dbcursor)
    dbcursor.execute(
        "SELECT count(*) from mcd where product_id = %s",
        (prod.get_product_id(),),
    )
    assert dbcursor.fetchone()[0] == 1


def test_170926_nodbinsert(dbcursor):
    """This product never hit the database for some reason?"""
    prod = parser(get_test_file("MCD_MPD/SWOMCD_2010.txt"))
    prod.database_save(dbcursor)
    dbcursor.execute(
        """
        SELECT * from mcd where product_id = %s
    """,
        (prod.get_product_id(),),
    )
    assert dbcursor.rowcount == 1


def test_mpd_mcdparser(dbcursor):
    """ The mcdparser can do WPC's MPD as well, test it """
    prod = parser(get_test_file("MCD_MPD/MPD.txt"))
    assert abs(prod.geometry.area - 4.657) < 0.001
    assert prod.attn_wfo == ["PHI", "AKQ", "CTP", "LWX"]
    assert prod.attn_rfc == ["MARFC"]
    ans = (
        "#WPC issues MPD 98 concerning HEAVY RAINFALL: NRN VA...D.C"
        "....CENTRAL MD INTO SERN PA "
        "https://www.wpc.ncep.noaa.gov/metwatch/metwatch_mpd_multi.php"
        "?md=98&yr=2013"
    )
    assert prod.tweet() == ans
    ans = (
        "Weather Prediction Center issues "
        "Mesoscale Precipitation Discussion #98 concerning HEAVY RAINFALL"
        " https://www.wpc.ncep.noaa.gov/metwatch/metwatch_mpd_multi.php"
        "?md=98&amp;yr=2013"
    )
    assert prod.get_jabbers("http://localhost")[0][0] == ans
    prod.database_save(dbcursor)


def test_mcdparser(dbcursor):
    """ Test Parsing of MCD Product """
    prod = parser(get_test_file("MCD_MPD/SWOMCD.txt"))
    assert abs(prod.geometry.area - 4.302) < 0.001
    assert prod.discussion_num == 1525
    assert prod.attn_wfo[2] == "DLH"
    ans = "PORTIONS OF NRN WI AND THE UPPER PENINSULA OF MI"
    assert prod.areas_affected == ans

    # With probability this time
    prod = parser(get_test_file("MCD_MPD/SWOMCDprob.txt"))
    assert abs(prod.geometry.area - 2.444) < 0.001
    assert prod.watch_prob == 20

    jmsg = prod.get_jabbers("http://localhost")
    ans = (
        "<p>Storm Prediction Center issues "
        '<a href="https://www.spc.noaa.gov/'
        'products/md/2013/md1678.html">Mesoscale Discussion #1678</a> '
        "concerning SEVERE POTENTIAL [watch probability: 20%] "
        '(<a href="http://localhost'
        '?pid=201308091725-KWNS-ACUS11-SWOMCD">View text</a>)</p>'
    )
    assert jmsg[0][1] == ans
    ans = (
        "Storm Prediction Center issues Mesoscale Discussion #1678 "
        "concerning SEVERE POTENTIAL [watch probability: 20%] "
        "https://www.spc.noaa.gov/products/md/2013/md1678.html"
    )
    assert jmsg[0][0] == ans
    ans = utc(2013, 8, 9, 17, 25)
    assert prod.sts == ans
    ans = utc(2013, 8, 9, 19, 30)
    assert prod.ets == ans

    prod.database_save(dbcursor)



"""NWS Local Storm Report (LSR) Parsing."""
import datetime
import re
import math


from shapely.geometry import Point as ShapelyPoint
from pyiem.nws.product import TextProduct, TextProductException
from pyiem.nws.lsr import LSR
from pyiem.util import utc
from pyiem import reference

# Don't permit LSRs that are more than 1 hour newer than product time
# or future of the current time
FUTURE_THRESHOLD = datetime.timedelta(hours=1)
SPLITTER = re.compile(
    r"(^[0-9].+?\n^[0-9].+?\n)((?:.*?\n)+?)(?=^[0-9]|$)", re.MULTILINE
)


class LSRProductException(TextProductException):
    """ Something we can raise when bad things happen! """


class LSRProduct(TextProduct):
    """ Represents a text product of the LSR variety """

    def __init__(
        self, text, utcnow=None, ugc_provider=None, nwsli_provider=None
    ):
        """ constructor """
        self.lsrs = []
        self.duplicates = 0
        TextProduct.__init__(
            self,
            text,
            utcnow=utcnow,
            ugc_provider=ugc_provider,
            nwsli_provider=nwsli_provider,
        )

    def get_temporal_domain(self):
        """ Return the min and max timestamps of lsrs """
        if not self.lsrs:
            return None, None
        valids = []
        for lsr in self.lsrs:
            valids.append(lsr.valid)
        return min(valids), max(valids)

    def is_summary(self):
        """ Returns is this LSR is a summary or not """
        return self.unixtext.find("...SUMMARY") > 0

    def get_url(self, baseuri):
        """ Get the URL of this product """
        min_time, max_time = self.get_temporal_domain()
        wfo = self.source[1:]
        return "%s#%s/%s/%s" % (
            baseuri,
            wfo,
            min_time.strftime("%Y%m%d%H%M"),
            max_time.strftime("%Y%m%d%H%M"),
        )

    def get_jabbers(self, uri, _uri2=None):
        """ return a text and html variant for Jabber stuff """
        res = []
        if not self.lsrs:
            return res
        wfo = self.source[1:]
        url = self.get_url(uri)

        for mylsr in self.lsrs:
            if mylsr.duplicate:
                continue
            res.append(mylsr.get_jabbers(uri))

        if self.is_summary():
            extra_text = ""
            if self.duplicates > 0:
                extra_text = (
                    ", %s out of %s reports were previously "
                    "sent and not repeated here."
                ) % (self.duplicates, len(self.lsrs))
            text = "%s: %s issues Summary Local Storm Report %s %s" % (
                wfo,
                wfo,
                extra_text,
                url,
            )

            html = (
                "<p>%s issues "
                "<a href='%s'>Summary Local Storm Report</a>%s</p>"
            ) % (wfo, url, extra_text)
            xtra = {
                "product_id": self.get_product_id(),
                "channels": "LSR%s" % (wfo,),
            }
            res.append([text, html, xtra])
        return res


def _mylowercase(text):
    """ Specialized lowercase function """
    tokens = text.split()
    for i, t in enumerate(tokens):
        if len(t) > 3:
            tokens[i] = t.title()
        elif t in [
            "N",
            "NNE",
            "NNW",
            "NE",
            "E",
            "ENE",
            "ESE",
            "SE",
            "S",
            "SSE",
            "SSW",
            "SW",
            "W",
            "WSW",
            "WNW",
            "NW",
        ]:
            continue
    return " ".join(tokens)


def parse_lsr(prod, text):
    """Emit a LSR object based on this text!
    0914 PM     HAIL             SHAW                    33.60N 90.77W
    04/29/2005  1.00 INCH        BOLIVAR            MS   EMERGENCY MNGR
    """
    lines = text.split("\n")
    if len(lines) < 2:
        prod.warnings.append(
            ("LSR text is too short |%s|\n%s")
            % (text.replace("\n", "<NL>"), text)
        )
        return None
    lsr = LSR()
    lsr.product = prod
    lsr.text = text
    tokens = lines[0].split()
    h12 = tokens[0][:-2]
    mm = tokens[0][-2:]
    ampm = tokens[1]
    dstr = "%s:%s %s %s" % (h12, mm, ampm, lines[1][:10])
    lsr.valid = datetime.datetime.strptime(dstr, "%I:%M %p %m/%d/%Y")
    lsr.assign_timezone(prod.tz, prod.z)
    # Check that we are within bounds
    if lsr.utcvalid > (prod.valid + FUTURE_THRESHOLD) or lsr.utcvalid > (
        utc() + FUTURE_THRESHOLD
    ):
        prod.warnings.append(
            (
                "LSR is from the future!\n"
                "prod.valid: %s lsr.valid: %s\n"
                "%s\n"
            )
            % (prod.valid, lsr.valid, text)
        )
        return None

    lsr.wfo = prod.source[1:]

    lsr.typetext = lines[0][12:29].strip()
    if lsr.typetext.upper() not in reference.lsr_events:
        prod.warnings.append(
            ("Unknown lsr.typetext |%s|\n%s") % (lsr.typetext, text)
        )
        return None

    lsr.city = lines[0][29:53].strip()

    tokens = lines[0][53:].strip().split()
    lat = float(tokens[0][:-1])
    lon = 0 - float(tokens[1][:-1])
    if lon <= -180 or lon >= 180 or lat >= 90 or lat <= -90:
        prod.warnings.append(
            ("Invalid Geometry Lat: %s Lon: %s\n%s") % (lat, lon, text)
        )
        return None
    lsr.geometry = ShapelyPoint((lon, lat))

    lsr.consume_magnitude(lines[1][12:29].strip())
    if lsr.magnitude_f is not None and math.isnan(lsr.magnitude_f):
        prod.warnings.append("LSR has NAN magnitude\n%s" % (text,))
        return None
    lsr.county = lines[1][29:48].strip()
    lsr.state = lines[1][48:50]
    lsr.source = lines[1][53:].strip()
    if len(lines) > 2:
        meat = " ".join(lines[2:]).strip()
        if meat.strip() != "":
            lsr.remark = " ".join(meat.split())
    return lsr


def parser(text, utcnow=None, ugc_provider=None, nwsli_provider=None):
    """Helper function that actually converts the raw text and emits an
    LSRProduct instance or returns an exception"""
    prod = LSRProduct(
        text, utcnow, ugc_provider=ugc_provider, nwsli_provider=nwsli_provider
    )
    if prod.z is None:
        prod.warnings.append("Abort parsing as no timezone was found.")
        return prod
    for match in SPLITTER.finditer(prod.unixtext):
        lsr = parse_lsr(prod, "".join(match.groups()))
        if lsr is None:
            continue
        prod.lsrs.append(lsr)

    return prod



"""Test CLI products"""
# pylint: disable=redefined-outer-name
import datetime

import pytest
import psycopg2.extras
from pyiem.reference import TRACE_VALUE
from pyiem.util import utc, get_test_file, get_dbconn
from pyiem.nws.products import parser as cliparser
from pyiem.nws.products.cli import get_number, CLIException

NWSLI_PROVIDER = {
    "IAD": dict(name="HOUSTON INTERCONTINENTAL", network="ZZ_ASOS"),
    "DMH": dict(name="", network="ZZ_ASOS"),
    "HOU": dict(name="HOUSTON/HOBBY AIRPORT", network="ZZ_ASOS"),
    "PADQ": dict(name="KODIAK", network="AK_ASOS"),
    "PAKN": dict(name="KING SALMON", network="AK_ASOS"),
    "PANC": dict(name="ANCHORAGE AK", network="AK_ASOS"),
    "PASN": dict(name="SAINT PAUL ISLAND", network="AK_ASOS"),
    "PBET": dict(name="BETHEL", network="AK_ASOS"),
    "PCBD": dict(name="COLD BAY", network="AK_ASOS"),
    "POME": dict(name="NOME WSO AP", network="AK_ASOS"),
}


def factory(fn):
    """Common cliparser logic."""
    return cliparser(get_test_file(fn), nwsli_provider=NWSLI_PROVIDER)


@pytest.fixture
def dbcursor():
    """Get a database cursor for testing."""
    dbconn = get_dbconn("iem")
    yield dbconn.cursor(cursor_factory=psycopg2.extras.DictCursor)
    dbconn.close()


def test_wrong_wmo_header():
    """Test that we raise an exception for this."""
    prod = cliparser(
        get_test_file("CLI/CLIANC.txt").replace("CDAK48", "XXXX48")
    )
    assert prod.valid


def test_bad_diction():
    """Test that we raise an exception for this."""
    with pytest.raises(CLIException):
        factory("CLI/CLIPPG.txt")


def test_bad_headline():
    """Test that we raise an exception for this."""
    with pytest.raises(CLIException):
        cliparser(
            get_test_file("CLI/CLIANC.txt").replace("WEATHER ITEM", "XXXX48"),
            nwsli_provider=NWSLI_PROVIDER,
        )


def test_bad_temperature():
    """Test that we raise an exception for this."""
    with pytest.raises(CLIException):
        cliparser(
            get_test_file("CLI/CLIANC.txt").replace(
                "TEMPERATURE", "TEMP3RATURE"
            ),
            nwsli_provider=NWSLI_PROVIDER,
        )


def test_no_weather_item():
    """Test that we raise an exception for this."""
    with pytest.raises(CLIException):
        cliparser(get_test_file("CLI/CLIHOU.txt").replace("WEATHER", "X"))


def test_unknown_station_in_multi():
    """Test that we raise an exception for this."""
    with pytest.raises(CLIException):
        cliparser(get_test_file("CLI/CLIHOU.txt"))


def test_parse_temperature_bad_token():
    """Test when we observe a bad temperature token."""
    prod = cliparser(
        get_test_file("CLI/CLIDMH.txt").replace("AVERAGE  ", "AV3RAG3  ")
    )
    assert prod.data


def test_get_number():
    """Test the number to float conversion."""
    assert get_number("") is None
    assert get_number("ABC") is None


def test_190510_parsefail(dbcursor):
    """This CLIDMH is not happy."""
    # Create an entry to actually update
    dbcursor.execute(
        "INSERT into stations(iemid, id, network) VALUES (%s, %s, %s)",
        (-1, "DMH", "ZZ_ASOS"),
    )
    prod = factory("CLI/CLIDMH.txt")
    assert not prod.warnings
    assert prod.data[0]["access_network"] == NWSLI_PROVIDER["DMH"]["network"]
    assert prod.data[0]["data"]["temperature_maximum"] == 74
    prod.sql(dbcursor)
    # Run twice to hit a no-op
    prod = factory("CLI/CLIDMH.txt")
    prod.sql(dbcursor)
    assert not prod.warnings


def test_200423_missing_skycover():
    """Test that we are processing skycover properly."""
    prod = factory("CLI/CLICVG.txt")
    assert prod.data[0]["data"]["average_sky_cover"] == 0.4


def test_database_progression(dbcursor):
    """Test our deletion logic."""

    def _get():
        """Fetch our current value."""
        dbcursor.execute(
            "SELECT high from cli_data "
            "where station = 'KCVG' and valid = '2020-04-22'"
        )
        return dbcursor.fetchone()["high"]

    prod = factory("CLI/CLICVG.txt")
    prod.sql(dbcursor)
    assert abs(_get() - 69.0) < 0.01
    prod = factory("CLI/CLICVG_older.txt")
    prod.sql(dbcursor)
    assert abs(_get() - 69.0) < 0.01
    prod = factory("CLI/CLICVG_newer.txt")
    prod.sql(dbcursor)
    assert abs(_get() - 70.0) < 0.01


def test_issue15_wind(dbcursor):
    """Test parsing of available wind information."""
    prod = factory("CLI/CLICVG.txt")
    assert abs(prod.data[0]["data"]["resultant_wind_speed"] - 6.0) < 0.01
    assert abs(prod.data[0]["data"]["resultant_wind_direction"] - 180.0) < 0.01
    assert abs(prod.data[0]["data"]["highest_wind_speed"] - 20.0) < 0.01
    assert abs(prod.data[0]["data"]["highest_wind_direction"] - 220.0) < 0.01
    assert abs(prod.data[0]["data"]["highest_gust_speed"] - 26.0) < 0.01
    assert abs(prod.data[0]["data"]["highest_gust_direction"] - 230.0) < 0.01
    assert abs(prod.data[0]["data"]["average_wind_speed"] - 7.7) < 0.01
    prod.sql(dbcursor)
    assert prod.data[0]["db_station"] == "KCVG"


def test_180208_issue56_tweetmissing():
    """Report None values as missing, not None"""
    prod = factory("CLI/CLIFFC.txt")
    j = prod.get_jabbers("http://localhost", "http://localhost")
    ans = (
        "PEACHTREE CITY Oct 3 Climate: Hi: 79 Lo: 67 Precip: 0.87 "
        "Snow: Missing "
        "http://localhost?pid=201410032032-KFFC-CDUS42-CLIFFC"
    )
    assert j[0][2]["twitter"] == ans


def test_170530_none():
    """CLILWD errored in production, so we add a test!"""
    prod = factory("CLI/CLILWD.txt")
    assert prod.data[0]["data"]["temperature_maximum"] == 76


def test_170315_invalid_dup():
    """CLIANC incorrectly has two CLIs"""
    prod = factory("CLI/CLIANC.txt")
    answers = [23, 22, 31, 10, 29, 33]
    for i, answer in enumerate(answers):
        assert prod.data[i]["data"]["temperature_maximum"] == answer


def test_151019_clibna():
    """CLIBNA is a new diction"""
    prod = factory("CLI/CLIBNA.txt")
    assert prod.data[0]["data"]["temperature_maximum"] == 47


def test_150303_alaska():
    """CLIANN Attempt to account for the badly formatted CLIs"""
    prod = factory("CLI/CLIANN.txt")
    assert prod.data[0]["data"]["temperature_maximum_time"] == "0151 PM"


def test_150112_climso():
    """ CLIMSO_2 found some issue with this in production? """
    prod = factory("CLI/CLIMSO_2.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 16


def test_141230_newregime4():
    """ CLIMBS has a new regime """
    prod = factory("CLI/CLIMBS.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 18


def test_141230_newregime3():
    """ CLICKV has a new regime """
    prod = factory("CLI/CLICKV.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 33


def test_141230_newregime2():
    """ CLISEW has a new regime """
    prod = factory("CLI/CLISEW.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 34


def test_141230_newregime():
    """ CLITCS has a new regime """
    prod = factory("CLI/CLITCS.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 22


def test_141229_newregime9():
    """ CLIMAI has a new regime """
    prod = factory("CLI/CLIMAI.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 61


def test_141229_newregime8():
    """ CLIECP has a new regime """
    prod = factory("CLI/CLIECP.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 62


def test_141229_newregime7():
    """ CLIBOI has a new regime """
    prod = factory("CLI/CLIBOI.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 23


def test_141229_newregime6():
    """ CLIMSO has a new regime """
    prod = factory("CLI/CLIMSO.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 12


def test_141229_newregime4():
    """ CLIOLF has a new regime """
    prod = factory("CLI/CLIOLF.txt")
    assert prod.data[0]["data"]["temperature_average"] == -2


def test_141229_newregime5():
    """ CLIICT has a new regime """
    prod = factory("CLI/CLIICT.txt")
    assert prod.data[0]["data"]["precip_today"] == 0.00


def test_141229_newregime3():
    """ CLIDRT has a new regime """
    prod = factory("CLI/CLIDRT.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 37


def test_141229_newregime2():
    """ CLIFMY has a new regime """
    prod = factory("CLI/CLIFMY.txt")
    assert prod.data[0]["data"]["temperature_minimum"] == 63


def test_141229_newregime():
    """ CLIEKA has a new regime """
    prod = factory("CLI/CLIEKA.txt")
    assert prod.data[0]["data"]["precip_today_record_years"][0] == 1896
    assert prod.data[0]["data"]["precip_today_record_years"][1] == 1999


def test_141215_convkey():
    """ CLIACT Get a warning about convert key """
    prod = factory("CLI/CLIACT.txt")
    assert prod.data[0]["data"]["snow_today_record_years"][0] == 1947
    assert prod.data[0]["data"]["snow_today_record_years"][1] == 1925


def test_141201_clihou():
    """ CLIHOU See that we can finally parse the CLIHOU product! """
    prod = factory("CLI/CLIHOU.txt")
    assert prod.data[0]["cli_station"] == "HOUSTON INTERCONTINENTAL"
    assert prod.data[1]["cli_station"] == "HOUSTON/HOBBY AIRPORT"


def test_141114_coopaux():
    """ CLIEAR Product had aux COOP data, which confused ingest """
    prod = factory("CLI/CLIEAR.txt")
    assert prod.data[0]["data"]["precip_today"] == 0


def test_141103_recordsnow():
    """ CLIBGR Make sure we can deal with record snowfall, again... """
    prod = factory("CLI/CLIBGR.txt")
    assert prod.data[0]["data"]["snow_today"] == 12.0


def test_141024_recordsnow(dbcursor):
    """ CLIOME See that we can handle record snowfall """
    prod = factory("CLI/CLIOME.txt")
    assert prod.data[0]["data"]["snow_today"] == 3.6
    prod.sql(dbcursor)


def test_141022_correction():
    """ CLIEWN See what happens if we have a valid product correction """
    prod = factory("CLI/CLIEWN.txt")
    assert prod.data[0]["data"]["temperature_maximum"] == 83


def test_141013_missing():
    """ CLIEST See why Esterville was not going to the database! """
    prod = factory("CLI/CLIEST.txt")
    assert prod.data[0]["data"]["temperature_maximum"] == 62
    assert prod.data[0]["data"]["precip_month"] == 1.22


def test_141013_tracetweet():
    """ CLIDSM2 Make sure we convert trace amounts in tweet to trace! """
    prod = factory("CLI/CLIDSM2.txt")
    j = prod.get_jabbers("http://localhost", "http://localhost")
    ans = (
        "DES MOINES IA Oct 12 Climate: Hi: 56 "
        "Lo: 43 Precip: Trace Snow: 0.0"
        " http://localhost?pid=201410122226-KDMX-CDUS43-CLIDSM"
    )
    assert j[0][2]["twitter"] == ans


def test_141003_missing():
    """ CLIFFC We are missing some data! """
    prod = factory("CLI/CLIFFC.txt")
    assert prod.data[0]["data"]["temperature_maximum_normal"] == 78


def test_141003_alaska():
    """ CLIBET Some alaska data was not getting processed"""
    prod = factory("CLI/CLIBET.txt")
    assert prod.data[0]["data"]["temperature_maximum"] == 17
    assert prod.data[0]["data"]["snow_jul1"] == 14.4


def test_140930_negative_temps():
    """ CLIALO Royal screwup not supporting negative numbers """
    prod = factory("CLI/CLIALO.txt")
    assert prod.data[0]["data"].get("temperature_minimum") == -21
    assert prod.data[0]["data"].get("temperature_minimum_record") == -21
    assert prod.data[0]["data"].get("snow_today") == 0.0
    assert prod.data[0]["data"].get("snow_today_record") == 13.2
    assert prod.data[0]["data"].get("snow_today_last") == 0.0
    assert prod.data[0]["data"].get("snow_month_last") == TRACE_VALUE
    assert prod.data[0]["data"].get("snow_jul1_last") == 11.3
    assert prod.data[0]["data"].get("average_sky_cover") is None


def test_140930_mm_precip():
    """ CLIABY Make sure having MM as today's precip does not error out """
    prod = factory("CLI/CLIABY.txt")
    assert prod.data[0]["data"].get("precip_today") is None


def test_cli():
    """ CLIJUN Test the processing of a CLI product """
    prod = factory("CLI/CLIJNU.txt")
    assert prod.data[0]["cli_valid"] == datetime.date(2013, 6, 30)
    assert prod.valid == utc(2013, 7, 1, 0, 36)
    assert prod.data[0]["data"]["temperature_maximum"] == 75
    assert prod.data[0]["data"]["temperature_maximum_time"] == "259 PM"
    assert prod.data[0]["data"]["temperature_minimum_time"] == "431 AM"
    assert prod.data[0]["data"]["precip_today"] == TRACE_VALUE

    j = prod.get_jabbers("http://localhost")
    ans = (
        "JUNEAU Jun 30 Climate Report: High: 75 "
        "Low: 52 Precip: Trace Snow: M "
        "http://localhost?pid=201307010036-PAJK-CDAK47-CLIJNU"
    )
    assert j[0][0] == ans


def test_cli2():
    """ CLIDSM test """
    prod = factory("CLI/CLIDSM.txt")
    assert prod.data[0]["cli_valid"] == datetime.date(2013, 8, 1)
    assert prod.data[0]["data"]["temperature_maximum"] == 89
    assert prod.data[0]["data"]["snow_month"] == 0
    assert prod.data[0]["data"]["temperature_minimum_record_years"][0] == 1898
    assert prod.data[0]["data"]["snow_today"] == 0
    assert prod.data[0]["data"]["precip_jun1"] == 4.25
    assert prod.data[0]["data"]["precip_jan1"] == 22.56


def test_cli3():
    """ CLINYC test """
    prod = factory("CLI/CLINYC.txt")
    assert prod.data[0]["data"]["snow_today_record_years"][0] == 1925
    assert prod.data[0]["data"]["snow_today_record"] == 11.5
    assert abs(prod.data[0]["data"]["average_sky_cover"] - 0.0) < 0.01



"""
  Classes Representing various variables
"""
# pylint: disable=no-member
import numpy as np
from pyiem.exceptions import UnitsError


class basetype:
    """ Base Object for all our vars """

    def __init__(self, value, units):
        """ constructor with value and units required """
        if units.upper() not in self.known_units:
            raise UnitsError(
                "unrecognized temperature unit: %s known: %s"
                % (units, self.known_units)
            )
        self._units = units.upper()
        if isinstance(value, list):
            self._value = np.ma.array(value, mask=[v is None for v in value])
        else:
            self._value = value

    def get_units(self):
        """Returns the units for this current datatype

        Returns:
          (str): the unit of this datatype
        """
        return self._units


class mixingratio(basetype):
    """Mixing Ratio."""

    known_units = ["KG/KG"]

    def value(self, units):
        """ Convert the value into the provided units """
        if units.upper() not in mixingratio.known_units:
            raise UnitsError(
                "unrecognized mixingratio unit: %s known: %s"
                % (units, mixingratio.known_units)
            )
        if units.upper() == self._units:
            return self._value


class distance(basetype):
    """ Distance """

    known_units = ["SM", "MI", "M", "KM", "FT", "MM", "IN", "CM"]

    def value(self, units):
        """ Convert to a value in the given units """
        units = units.upper()
        if units not in distance.known_units:
            raise UnitsError(
                "unrecognized distance unit: %s known: %s"
                % (units, distance.known_units)
            )
        if units == self._units:
            return self._value

        # MKS
        if self._units in ["SM", "MI"]:
            meters = self._value * 1609.344
        elif self._units == "FT":
            meters = self._value / 3.28084
        elif self._units == "KM":
            meters = self._value * 1000.0
        elif self._units == "MM":
            meters = self._value / 1000.0
        elif self._units == "CM":
            meters = self._value / 100.0
        elif self._units == "IN":
            meters = self._value * 0.0254
        else:
            meters = self._value
        # Output
        if units in ["SM", "MI"]:
            return meters / 1609.344
        elif units == "FT":
            return meters * 3.28084
        elif units == "KM":
            return meters / 1000.0
        elif units == "MM":
            return meters * 1000.0
        elif units == "CM":
            return meters * 100.0
        elif units == "IN":
            return meters / 0.0254
        return meters


class precipitation(basetype):
    """ Precipitation """

    known_units = ["IN", "CM", "MM"]

    def value(self, units):
        """ Convert to a value in the given units """
        units = units.upper()
        if units not in precipitation.known_units:
            raise UnitsError(
                ("unrecognized precipitation unit: %s known: %s")
                % (units, precipitation.known_units)
            )
        if units == self._units:
            return self._value

        # MKS
        if self._units == "IN":
            mm = self._value * 25.4
        elif self._units == "CM":
            mm = self._value * 10.0
        else:
            mm = self._value

        # Convert
        if units == "MM":
            return mm
        elif units == "CM":
            return mm * 10.0
        elif units == "IN":
            return mm / 25.4


class speed(basetype):
    """ Speed """

    known_units = ["KT", "MPH", "MPS", "KMH", "KTS"]

    def value(self, units=None):
        """ Convert to a value in the given units """
        if units is None:
            units = self._units
        units = units.upper()
        if units not in speed.known_units:
            raise UnitsError(
                "unrecognized speed unit: %s known: %s"
                % (units, speed.known_units)
            )
        if units == self._units:
            return self._value
        # MKS
        if self._units == "KMH":
            mps_value = self._value / 3.6
        elif self._units in ["KT", "KTS"]:
            mps_value = self._value * 0.514444
        elif self._units == "MPH":
            mps_value = self._value * 0.447000
        else:
            mps_value = self._value
        # return
        if units == "KMH":
            return mps_value * 3.6
        elif units in ["KT", "KTS"]:
            return mps_value / 0.514444
        elif units == "MPH":
            return mps_value / 0.447000
        elif units == "MPS":
            return mps_value


class humidity(basetype):
    """ Humidity, this is not as straight forward as the others """

    known_units = ["%"]

    def value(self, units):
        """ Convert to a value in the given units """
        if units.upper() not in humidity.known_units:
            raise UnitsError(
                "unrecognized humidity unit: %s known: %s"
                % (units, humidity.known_units)
            )
        return self._value


class direction(basetype):
    """ Direction from North """

    known_units = ["DEG", "RAD"]

    def value(self, units):
        """ Convert to a value in the given units """
        if units.upper() not in direction.known_units:
            raise UnitsError(
                "unrecognized direction unit: %s known: %s"
                % (units, direction.known_units)
            )
        if units.upper() == self._units:
            return self._value

        if self._units == "DEG" and units.upper() == "RAD":
            return self._value * 3.1415926535897931 / 180.0
        if self._units == "RAD" and units.upper() == "DEG":
            return self._value * 180 / 3.1415926535897931


class pressure(basetype):
    """ Pressure """

    known_units = ["MB", "HPA", "IN", "PA"]

    def value(self, units):
        """ Convert to a value in the given units """
        if units.upper() not in pressure.known_units:
            raise UnitsError(
                "unrecognized pressure unit: %s known: %s"
                % (units, pressure.known_units)
            )
        if units.upper() == self._units:
            return self._value

        # MKS
        if self._units == "IN":
            mb_value = self._value * 33.86398
        elif self._units == "PA":
            mb_value = self._value / 100.0
        else:
            mb_value = self._value
        # Now convert
        if units.upper() in ["MB", "HPA"]:
            return mb_value
        elif units.upper() == "PA":
            return mb_value * 100.0
        elif units.upper() == "IN":
            return mb_value / 33.86398


class temperature(basetype):
    """ Temperature """

    known_units = ["F", "K", "C"]

    def value(self, units):
        """ Convert to a value in the given units """
        units = units.upper()
        if units not in temperature.known_units:
            raise UnitsError(
                "unrecognized temperature unit: %s known: %s"
                % (units, temperature.known_units)
            )
        if units == self._units:
            return self._value

        # Convert to Celsius first
        if self._units == "C":
            celsius_value = self._value
        elif self._units == "F":
            celsius_value = (self._value - 32.0) / 1.8
        elif self._units == "K":
            celsius_value = self._value - 273.15
        # Dump back
        if units == "C":
            return celsius_value
        elif units == "K":
            return 273.15 + celsius_value
        elif units == "F":
            return 32.0 + celsius_value * 1.8



"""Base Class encapsulating a NWS Text Product"""
from datetime import timezone, timedelta, datetime
from collections import OrderedDict
import re

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

from shapely.geometry import Polygon, MultiPolygon
from shapely.wkt import dumps

from pyiem import reference
from pyiem.util import LOG
from pyiem.exceptions import TextProductException
from pyiem.nws import ugc, vtec, hvtec


# The AWIPS Product Identifier is supposed to be 6chars as per directive,
# but in practice it is sometimes something between 4 and 6 chars
# We do require that the first character be a A-Z one as otherwise this will
# match the LDM sequence number at the top!
AFOSRE = re.compile(r"^([A-Z][A-Z0-9]{3,5})\s*\t*$", re.M)
TIME_FMT = (
    "([0-9:]+) (AM|PM) ([A-Z][A-Z][A-Z]?T) ([A-Z][A-Z][A-Z]) "
    "([A-Z][A-Z][A-Z]) ([0-9]+) ([1-2][0-9][0-9][0-9])"
)
TIME_RE = re.compile(f"^{TIME_FMT}$", re.M | re.IGNORECASE)
# Sometimes products have a duplicated timestamp in another tz
TIME_EXT_RE = re.compile(
    rf"^{TIME_FMT}\s?/\s?{TIME_FMT}\s?/$", re.M | re.IGNORECASE
)
# Without the line start and end requirement
TIME_RE_ANYWHERE = re.compile(f"{TIME_FMT}", re.IGNORECASE)

# Note that bbb of RTD is supported here, but does not appear to be allowed
WMO_RE = re.compile(
    "^(?P<ttaaii>[A-Z0-9]{4,6}) (?P<cccc>[A-Z]{4}) "
    r"(?P<ddhhmm>[0-3][0-9][0-2][0-9][0-5][0-9])\s*"
    r"(?P<bbb>[ACR][ACMORT][A-Z])?\s*$",
    re.M,
)
TIME_MOT_LOC = re.compile(
    r"TIME\.\.\.MOT\.\.\.LOC\s+(?P<ztime>[0-9]{4})Z\s+"
    r"(?P<dir>[0-9]{1,3})DEG\s+"
    r"(?P<sknt>[0-9]{1,3})KT\s+(?P<loc>[0-9 ]+)"
)
LAT_LON_PREFIX = re.compile(r"LAT\.\.\.LON", re.IGNORECASE)
LAT_LON = re.compile(r"([0-9]{4,8})\s+")
# This is a legacy tag that is no longer used, but we want to continue to
# parse it.
WINDHAIL = re.compile(
    r".*WIND\.\.\.HAIL (?P<winddir>[><]?)(?P<wind>[0-9]+)"
    "(?P<windunits>MPH|KTS) "
    r"(?P<haildir>[><]?)(?P<hail>[0-9\.]+)IN"
)
HAILTAG = re.compile(r".*HAIL\.\.\.(?P<haildir>[><]?)(?P<hail>[0-9\.]+)IN")
WINDTAG = re.compile(
    r".*WIND\.\.\.(?P<winddir>[><]?)\s?(?P<wind>[0-9]+)\s?"
    "(?P<windunits>MPH|KTS)"
)
TORNADOTAG = re.compile(
    r".*TORNADO\.\.\.(?P<tornado>RADAR INDICATED|" "OBSERVED|POSSIBLE)"
)
WATERSPOUTTAG = re.compile(
    r".*WATERSPOUT\.\.\.(?P<waterspout>RADAR INDICATED|OBSERVED|POSSIBLE)"
)
TORNADODAMAGETAG = re.compile(
    r".*TORNADO DAMAGE THREAT\.\.\."
    "(?P<damage>CONSIDERABLE|SIGNIFICANT|CATASTROPHIC)"
)
FLOOD_TAGS = re.compile(
    r".*(?P<key>FLASH FLOOD|FLASH FLOOD DAMAGE THREAT|EXPECTED RAINFALL|"
    r"DAM FAILURE|LEVEE FAILURE)\.\.\.(?P<value>.*?)\n"
)
TORNADO = re.compile(r"^AT |^\* AT")
RESENT = re.compile(r"\.\.\.(RESENT|RETRANSMITTED|CORRECTED)")
EMERGENCY_RE = re.compile(r"(TORNADO|FLASH\s+FLOOD)\s+EMERGENCY", re.I)
PDS_RE = re.compile(
    r"THIS\s+IS\s+A\s+PARTICULARLY\s+DANGEROUS\s+SITUATION", re.I
)

# http://www.nws.noaa.gov/os/notification/pns11mixedcase.txt
# DISALLOWED_CHARS = re.compile(r'[^\x40-\x7F]')
KNOWN_BAD_TTAAII = ["KAWN"]


def checker(lon, lat, strdata):
    """make sure our values are within physical bounds"""
    if lat >= 90 or lat <= -90:
        raise TextProductException(f"invalid latitude {lat} from {strdata}")
    if lon > 180 or lon < -180:
        raise TextProductException(f"invalid longitude {lon} from {strdata}")
    return (lon, lat)


def str2polygon(strdata):
    """Convert some string data into a polygon"""
    pts = []
    partial = None

    # We have two potential formats, one with 4 or 5 places and one
    # with eight!
    vals = re.findall(LAT_LON, strdata)
    for val in vals:
        if len(val) == 8:
            lat = float(val[:4]) / 100.00
            lon = float(val[4:]) / 100.00
            if lon < 40:
                lon += 100.0
            lon = 0 - lon
            pts.append(checker(lon, lat, strdata))
        else:
            fval = float(val) / 100.00
            if partial is None:  # we have lat
                partial = fval
                continue
            # we have a lon
            if fval < 40:
                fval += 100.0
            fval = 0 - fval
            pts.append(checker(fval, partial, strdata))
            partial = None

    if not pts:
        return None
    if pts[0][0] != pts[-1][0] and pts[0][1] != pts[-1][1]:
        pts.append(pts[0])
    return Polygon(pts)


def date_tokens2datetime(tokens):
    """Convert tokens from MND regex to a valid time, if possible.

    Returns:
      z (str): 3-4 char timezone string
      tz (datetime.timezone): of this product
      utcvalid (datetimetz): of this product
    """
    z = tokens[2].upper()
    tz = ZoneInfo(reference.name2pytz.get(z, "UTC"))
    hhmi = tokens[0]
    # False positive from regex
    if hhmi[0] == ":":
        hhmi = hhmi.replace(":", "")
    if hhmi.find(":") > -1:
        (hh, mi) = hhmi.split(":")
    elif len(hhmi) < 3:
        hh = hhmi
        mi = 0
    else:
        hh = hhmi[:-2]
        mi = hhmi[-2:]
    # Workaround 24 hour clock abuse
    if int(hh) > 12 and tokens[1].upper() == "PM":
        hh = int(hh) - 12
    dstr = "%s:%s %s %s %s %s" % (
        hh,
        mi,
        tokens[1],
        tokens[4],
        tokens[5],
        tokens[6],
    )
    # Careful here, need to go to UTC time first then come back!
    now = datetime.strptime(dstr, "%I:%M %p %b %d %Y")
    now += timedelta(hours=reference.offsets.get(z, 0))
    return z, tz, now.replace(tzinfo=timezone.utc)


class TextProductSegment:
    """ A segment of a Text Product """

    def __init__(self, text, tp):
        """ Constructor """
        self.unixtext = text
        self.tp = tp  # Reference to parent
        self.ugcs, self.ugcexpire = ugc.parse(
            text, tp.valid, ugc_provider=tp.ugc_provider
        )
        self.vtec = vtec.parse(text)
        self.headlines = self.parse_headlines()
        self.hvtec = hvtec.parse(text, nwsli_provider=tp.nwsli_provider)

        # TIME...MOT...LOC Stuff!
        self.tml_giswkt = None
        self.tml_valid = None
        self.tml_sknt = None
        self.tml_dir = None
        self.process_time_mot_loc()

        #
        self.giswkt = None
        self.sbw = self.process_latlon()

        # tags
        self.windtag = None
        self.windtagunits = None
        self.hailtag = None
        self.haildirtag = None
        self.winddirtag = None
        self.tornadotag = None
        self.waterspouttag = None
        self.tornadodamagetag = None
        # allows for deterministic testing of results
        self.flood_tags = OrderedDict()
        self.is_emergency = False
        self.is_pds = False
        self.process_tags()

        self.bullets = self.process_bullets()

    def get_ugcs_tuple(self):
        """ Helper to return a tuple useful for SQL. """
        return tuple([str(u) for u in self.ugcs])

    def get_hvtec_nwsli(self):
        """ Return the first hvtec NWSLI entry, if it exists """
        if not self.hvtec:
            return None
        return self.hvtec[0].nwsli.id

    def get_hvtec_cause(self):
        """ Return the first hvtec cause entry, if it exists """
        if not self.hvtec:
            return None
        return self.hvtec[0].cause

    def get_hvtec_severity(self):
        """ Return the first hvtec severity entry, if it exists """
        if not self.hvtec:
            return None
        return self.hvtec[0].severity

    def get_hvtec_record(self):
        """ Return the first hvtec record entry, if it exists """
        if not self.hvtec:
            return None
        return self.hvtec[0].record

    def svs_search(self):
        """ Special search the product for special text """
        sections = self.unixtext.split("\n\n")
        for section in sections:
            if TORNADO.findall(section):
                return " ".join(section.replace("\n", " ").split())
        return ""

    def process_bullets(self):
        """ Figure out the bulleted segments """
        parts = re.findall(r"^\*([^\*]*)", self.unixtext, re.M | re.DOTALL)
        bullets = []
        for part in parts:
            pos = part.find("\n\n")
            if pos > 0:
                bullets.append(" ".join(part[:pos].replace("\n", "").split()))
            else:
                bullets.append(" ".join(part.replace("\n", "").split()))
        return bullets

    def process_tags(self):
        """ Find various tags in this segment """
        nolf = self.unixtext.replace("\n", " ")
        res = EMERGENCY_RE.findall(nolf)
        if res:
            # TODO: this can be based off the IBW Tags too
            self.is_emergency = True
        res = PDS_RE.findall(nolf)
        if res:
            self.is_pds = True
        match = WINDHAIL.match(nolf)
        if match:
            gdict = match.groupdict()
            self.windtag = gdict["wind"]
            self.windtagunits = gdict["windunits"]
            self.haildirtag = gdict["haildir"]
            self.winddirtag = gdict["winddir"]
            self.hailtag = gdict["hail"]

        match = WINDTAG.match(nolf)
        if match:
            gdict = match.groupdict()
            self.winddirtag = gdict["winddir"]
            self.windtag = gdict["wind"]
            self.windtagunits = gdict["windunits"]

        match = HAILTAG.match(nolf)
        if match:
            gdict = match.groupdict()
            self.haildirtag = gdict["haildir"]
            self.hailtag = gdict["hail"]

        match = TORNADOTAG.match(nolf)
        if match:
            gdict = match.groupdict()
            self.tornadotag = gdict["tornado"]

        match = TORNADODAMAGETAG.match(nolf)
        if match:
            gdict = match.groupdict()
            self.tornadodamagetag = gdict["damage"]

        match = WATERSPOUTTAG.match(nolf)
        if match:
            gdict = match.groupdict()
            self.waterspouttag = gdict["waterspout"]

        for token in FLOOD_TAGS.findall(self.unixtext):
            self.flood_tags[token[0]] = token[1]

    def special_tags_to_text(self):
        """
        Convert the special tags into a nice text
        """
        if all(
            [
                self.windtag is None,
                self.tornadotag is None,
                self.hailtag is None,
                self.tornadodamagetag is None,
                self.waterspouttag is None,
                not self.flood_tags,
            ]
        ):
            return ""

        parts = []
        if self.tornadotag is not None:
            parts.append("tornado: %s" % (self.tornadotag))
        if self.waterspouttag is not None:
            parts.append("waterspout: %s" % (self.waterspouttag))
        if self.tornadodamagetag is not None:
            parts.append("tornado damage threat: %s" % (self.tornadodamagetag))
        if self.windtag is not None:
            parts.append(
                "wind: %s%s %s"
                % (
                    self.winddirtag.replace(">", "&gt;").replace("<", "&lt;"),
                    self.windtag,
                    self.windtagunits,
                )
            )
        if self.hailtag is not None:
            parts.append(
                "hail: %s%s IN"
                % (
                    self.haildirtag.replace(">", "&gt;").replace("<", "&lt;"),
                    self.hailtag,
                )
            )
        for k, v in self.flood_tags.items():
            parts.append("%s: %s" % (k.lower(), v.lower()))
        return " [" + ", ".join(parts) + "] "

    def process_latlon(self):
        """Parse the segment looking for the 'standard' LAT...LON encoding"""
        data = self.unixtext.replace("\n", " ")
        search = LAT_LON_PREFIX.search(data)
        if search is None:
            return None
        pos = search.start()
        newdata = data[pos + 9 :]
        # Go find our next non-digit, non-space character, if we find it, we
        # should truncate our string, this could be improved, I suspect
        search = re.search(r"[^\s0-9]", newdata)
        if search is not None:
            pos2 = search.start()
            newdata = newdata[:pos2]

        poly = str2polygon(newdata)
        if poly is None:
            return None

        # check 0, PGUM polygons are east longitude akrherz/pyIEM#74
        if self.tp.source == "PGUM":
            newpts = [[0 - pt[0], pt[1]] for pt in poly.exterior.coords]
            poly = Polygon(newpts)

        # check 1, is the polygon valid?
        if not poly.is_valid:
            self.tp.warnings.append(
                ("LAT...LON polygon is invalid!\n%s") % (poly.exterior.xy,)
            )
            return None
        # check 2, is the exterior ring of the polygon clockwise?
        if poly.exterior.is_ccw:
            # No longer a warning as it was too much noise
            LOG.info(
                "LAT...LON polygon exterior is CCW, reversing\n%s",
                poly.exterior.xy,
            )
            poly = Polygon(
                zip(poly.exterior.xy[0][::-1], poly.exterior.xy[1][::-1])
            )
        self.giswkt = "SRID=4326;%s" % (
            dumps(MultiPolygon([poly]), rounding_precision=6),
        )
        return poly

    def process_time_mot_loc(self):
        """ Try to parse the TIME...MOT...LOC """
        pos = self.unixtext.find("TIME...MOT...LOC")
        if pos == -1:
            return
        if self.unixtext[pos - 1] != "\n":
            self.tp.warnings.append(
                "process_time_mot_loc segment likely has "
                "poorly formatted TIME...MOT...LOC"
            )
        search = TIME_MOT_LOC.search(self.unixtext)
        if not search:
            self.tp.warnings.append(
                "process_time_mot_loc segment find OK, but regex failed..."
            )
            return

        gdict = search.groupdict()
        if len(gdict["ztime"]) != 4 or self.ugcexpire is None:
            return
        hh = int(gdict["ztime"][:2])
        mi = int(gdict["ztime"][2:])
        self.tml_valid = self.ugcexpire.replace(hour=hh, minute=mi)
        if hh > self.ugcexpire.hour:
            self.tml_valid = self.tml_valid - timedelta(days=1)

        self.tml_valid = self.tml_valid.replace(tzinfo=timezone.utc)

        tokens = gdict["loc"].split()
        lats = []
        lons = []
        if len(tokens) % 2 != 0:
            tokens = tokens[:-1]
        if not tokens:
            return
        for i in range(0, len(tokens), 2):
            lats.append(float(tokens[i]) / 100.0)
            lons.append(0 - float(tokens[i + 1]) / 100.0)

        if len(lats) == 1:
            self.tml_giswkt = "SRID=4326;POINT(%s %s)" % (lons[0], lats[0])
        else:
            pairs = []
            for lat, lon in zip(lats, lons):
                pairs.append("%s %s" % (lon, lat))
            self.tml_giswkt = "SRID=4326;LINESTRING(%s)" % (",".join(pairs),)
        self.tml_sknt = int(gdict["sknt"])
        self.tml_dir = int(gdict["dir"])

    def parse_headlines(self):
        """ Find headlines in this segment """
        headlines = re.findall(
            r"^\.\.\.(.*?)\.\.\.[ ]?\n\n", self.unixtext, re.M | re.S
        )
        headlines = [
            " ".join(h.replace("...", ", ").replace("\n", " ").split())
            for h in headlines
        ]
        return headlines

    def get_affected_wfos(self):
        """Based on the ugc_provider, figure out which WFOs are impacted by
        this product segment"""
        affected_wfos = []
        for _ugc in self.ugcs:
            for wfo in _ugc.wfos:
                if wfo not in affected_wfos:
                    affected_wfos.append(wfo)

        return affected_wfos


class TextProduct:
    """class representing a NWS Text Product"""

    def __init__(
        self,
        text,
        utcnow=None,
        ugc_provider=None,
        nwsli_provider=None,
        parse_segments=True,
    ):
        """
        Constructor
        @param text string single text product
        @param utcnow used to compute offsets for when this product my be valid
        @param ugc_provider a dictionary of UGC objects already setup
        @param parse_segments should the segments be parsed as well? True
        """
        self.warnings = []

        self.text = text
        if ugc_provider is None:
            ugc_provider = {}
        if nwsli_provider is None:
            nwsli_provider = {}
        self.ugc_provider = ugc_provider
        self.nwsli_provider = nwsli_provider
        self.unixtext = text.replace("\r\r\n", "\n")
        self.sections = self.unixtext.split("\n\n")
        self.afos = None
        # The "truth" timestamp
        self.valid = None
        # The WMO header based timestamp
        self.wmo_valid = None
        self.source = None
        self.wmo = None
        self.ddhhmm = None
        self.bbb = None
        self.utcnow = utcnow
        self.segments = []
        self.z = None
        self.tz = None
        self.geometry = None
        if utcnow is None:
            self.utcnow = datetime.utcnow().replace(tzinfo=timezone.utc)
        # make sure this is actualing in UTC
        self.utcnow = self.utcnow.astimezone(timezone.utc)

        self.parse_wmo()
        self.parse_afos()
        self._parse_valid(utcnow)
        if parse_segments:
            self.parse_segments()

    def suv_iter(self):
        """Yield [(segment, ugcs, vtec)] combos found in product."""
        for segment in self.segments:
            if not segment.ugcs or not segment.vtec:
                continue
            for _vtec in segment.vtec:
                if _vtec.status == "T" or _vtec.action == "ROU":
                    continue
                yield (segment, segment.ugcs, _vtec)

    def is_resent(self):
        """ Check to see if this product is a ...RESENT product """
        return self.unixtext.find("...RESENT") > 0

    def is_correction(self):
        """Is this product a correction?

        Sadly, this is not as easy as it should be.  It turns out that some
        products do not have a proper correction mechanism, so offices will
        just brute force in a note into the MND header.  So we have to do
        some further checking...

        Returns:
          bool: Is this product a correction?
        """
        # OK, go looking for RESENT style tags, assume it happens within first
        # 300 chars
        if RESENT.search(self.text[:300]):
            return True
        if self.bbb is None or not self.bbb:
            return False
        if self.bbb[0] in ["A", "C"]:
            return True
        return False

    def get_channels(self):
        """ Return a list of channels """
        res = [self.afos, f"{self.afos[:3]}..."]
        if self.afos[:3] in ["TCU", "TCD", "TCM", "TCP", "TWO"]:
            res.append(self.afos[:5])
        return res

    def get_nicedate(self):
        """Nicely format the issuance time of this product"""
        if self.valid is None:
            return "(unknown issuance time)"
        localts = self.valid
        fmt = "%b %-d, %H:%M UTC"
        if self.tz is not None:
            localts = self.valid.astimezone(self.tz)
            # A bit of complexity as offices may not implement daylight saving
            if self.z.endswith("ST") and localts.dst():
                localts -= timedelta(hours=1)
            fmt = "%b %-d, %-I:%M %p " + self.z
        return localts.strftime(fmt)

    def get_main_headline(self, default=""):
        """Return a string for the main headline, if it exists"""
        for segment in self.segments:
            if segment.headlines:
                return segment.headlines[0]
        return default

    def get_jabbers(self, uri, _uri2=None):
        """Return a tuple of jabber messages [(plain, html, xtra_dict)]

        Args:
          uri (str): the base URI to use to construct links

        Returns:
          [(str, str, dict)]
        """
        templates = [
            "%(source)s issues %(name)s (%(aaa)s) at %(stamp)s%(headline)s ",
            "%(source)s issues %(name)s (%(aaa)s) at %(stamp)s ",
        ]
        aaa = self.afos[:3]
        hdl = self.get_main_headline()
        data = {
            "headline": " ...%s..." % (hdl,) if hdl != "" else "",
            "source": self.source[1:],
            "aaa": aaa,
            "name": reference.prodDefinitions.get(
                aaa, reference.prodDefinitions.get(self.afos, self.afos)
            ),
            "stamp": self.get_nicedate(),
            "url": "%s?pid=%s" % (uri, self.get_product_id()),
        }
        res = []
        plain = (templates[0] + "%(url)s") % data
        html = (
            '<p>%(source)s issues <a href="%(url)s">%(name)s (%(aaa)s)</a>'
            " at %(stamp)s</p>"
        ) % data
        tweet = templates[0] % data
        if (len(tweet) - 25) > reference.TWEET_CHARS:
            tweet = templates[1] % data
        tweet += data["url"]
        xtra = {
            "channels": ",".join(self.get_channels()),
            "product_id": self.get_product_id(),
            "twitter": tweet,
        }
        res.append((plain, html, xtra))
        return res

    def get_signature(self):
        """Find the signature at the bottom of the page"""
        return " ".join(
            self.segments[-1].unixtext.replace("\n", " ").strip().split()
        )

    def parse_segments(self):
        """ Split the product by its $$ """
        segs = self.unixtext.split("$$")
        for seg in segs:
            self.segments.append(TextProductSegment(seg, self))

    def get_product_id(self):
        """ Get an identifier of this product used by the IEM """
        pid = "%s-%s-%s-%s" % (
            self.valid.strftime("%Y%m%d%H%M"),
            self.source,
            self.wmo,
            self.afos,
        )
        return pid.strip()

    def _parse_valid(self, provided_utcnow):
        """Figure out the timestamp of this product.

        Args:
          provided_utcnow (datetime): What our library was provided for the UTC
            timestamp, it could be None
        """
        # The MND header hopefully has a full timestamp that is the best
        # truth that we can have for this product.
        tokens = TIME_RE.findall(self.unixtext)
        if not tokens:
            tokens = TIME_EXT_RE.findall(self.unixtext)
            if not tokens:
                tokens = TIME_RE_ANYWHERE.findall(self.unixtext)
        if provided_utcnow is None and tokens:
            try:
                z, _tz, valid = date_tokens2datetime(tokens[0])
                if z not in reference.offsets:
                    self.warnings.append(f"product timezone '{z}' unknown")
            except ValueError:
                msg = (
                    "Invalid timestamp [%s] found in product "
                    "[%s %s %s] header"
                ) % (" ".join(tokens[0]), self.wmo, self.source, self.afos)
                raise TextProductException(self.source[1:], msg)

            # Set the utcnow based on what we found by looking at the header
            self.utcnow = valid

        # Search out the WMO header, this had better always be there
        # We only care about the first hit in the file, searching from top
        # Take the first hit, ignore others
        wmo_day = int(self.ddhhmm[:2])
        wmo_hour = int(self.ddhhmm[2:4])
        wmo_minute = int(self.ddhhmm[4:])

        self.wmo_valid = self.utcnow.replace(
            hour=wmo_hour, minute=wmo_minute, second=0, microsecond=0
        )
        if wmo_day != self.utcnow.day:
            if wmo_day - self.utcnow.day == 1:  # Tomorrow
                self.wmo_valid = self.wmo_valid.replace(day=wmo_day)
            elif wmo_day > 25 and self.utcnow.day < 15:  # Previous month!
                self.wmo_valid = self.wmo_valid + timedelta(days=-10)
                self.wmo_valid = self.wmo_valid.replace(day=wmo_day)
            elif wmo_day < 5 and self.utcnow.day >= 15:  # next month
                self.wmo_valid = self.wmo_valid + timedelta(days=10)
                self.wmo_valid = self.wmo_valid.replace(day=wmo_day)
            else:
                self.wmo_valid = self.wmo_valid.replace(day=wmo_day)

        # we can do no better
        self.valid = self.wmo_valid

        # If we don't find anything, lets default to now, its the best
        if not tokens:
            return
        self.z, self.tz, self.valid = date_tokens2datetime(tokens[0])

    def parse_wmo(self):
        """ Parse things related to the WMO header"""
        search = WMO_RE.search(self.unixtext[:100])
        if search is None:
            raise TextProductException(
                f"FATAL: Could not parse WMO header! '{self.text[:100]}'"
            )
        gdict = search.groupdict()
        self.wmo = gdict["ttaaii"]
        self.source = gdict["cccc"]
        self.ddhhmm = gdict["ddhhmm"]
        self.bbb = gdict["bbb"]
        if len(self.wmo) == 4:
            # Don't whine about known problems
            if (
                self.source not in KNOWN_BAD_TTAAII
                and not self.source.startswith("S")
            ):
                self.warnings.append(
                    f"WMO ttaaii found four chars: {self.wmo} {self.source} "
                    "adding 00"
                )
            self.wmo += "00"

    def get_affected_wfos(self):
        """Based on the ugc_provider, figure out which WFOs are impacted by
        this product"""
        affected_wfos = []
        for segment in self.segments:
            for ugcs in segment.ugcs:
                for wfo in ugcs.wfos:
                    if wfo not in affected_wfos:
                        affected_wfos.append(wfo)

        return affected_wfos

    def parse_afos(self):
        """ Figure out what the AFOS PIL is """
        # at most, only look at the top four lines
        data = "\n".join(
            [line.strip() for line in self.sections[0].split("\n")[:4]]
        )
        tokens = re.findall(r"^([A-Z0-9\s\t]{4,6})$", data, re.M)
        if tokens:
            self.afos = tokens[0].strip()



"""Our default IEM template environ."""
import datetime

from jinja2 import Environment, PackageLoader


# Can not support auto_escape at this time as parts of the template are
# being provided verbatim.
TEMPLATE_ENV = Environment(loader=PackageLoader("pyiem", "templates/iem"))


def get_template(filename):
    """Helper for getting a template.

    Args:
      filename (str): the template file to open.
    """
    tpl = TEMPLATE_ENV.get_template(filename)
    tpl.globals["footer_year"] = datetime.date.today().year
    return tpl


# Default template used downstream.
TEMPLATE = get_template("full.j2")



"""test cscap_utils."""
import tempfile
import os

from pyiem import cscap_utils as csu


def test_cleanvalue():
    """ see what we can do with cleaning strings"""
    assert abs(10.54 - csu.cleanvalue("10.54%")) < 0.01
    assert csu.cleanvalue("Did NOt Collect") == "did not collect"
    assert csu.cleanvalue("<0.2") == "< 0.2"
    assert csu.cleanvalue(" ") is None
    assert csu.cleanvalue("NA") == "n/a"
    assert abs(csu.cleanvalue("10") - 10.0) < 0.0001
    csu.cleanvalue("QQQ")
    assert "QQQ" in csu.CLEANVALUE_COMPLAINED


def test_missing_config():
    """Test that we can deal with a missing file."""
    with tempfile.NamedTemporaryFile() as tmp:
        tmpfn = tmp.name
    csu.CONFIG_FN = tmpfn
    cfg = csu.get_config()
    assert cfg is None
    csu.save_config({})


def test_config():
    """Make sure we exercise the config logic as things get hairy"""
    (_, tmpfn) = tempfile.mkstemp()
    # create bogus config file
    cfg = dict(a="a", b="b")
    # Write config to bogus file
    csu.save_config(cfg, tmpfn)
    # Attempt to load it back now
    cfg = csu.get_config(tmpfn)
    assert cfg is not None
    os.unlink(tmpfn)


def test_translateyears():
    """See that we can translate years properly"""
    x = csu.translate_years("X ('07-'17)")
    assert x[0] == 2007
    x = csu.translate_years("X ('98-'06)")
    assert x[0] == 1998
    assert x[-1] == 2006
    x = csu.translate_years("X ('14, '15, '16, '17)")
    assert x[0] == 2014
    assert x[-1] == 2017
    x = csu.translate_years("X ('06)")
    assert x[0] == 2006
